! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: confusion
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to do with calculating measures of source confusion.
! 
! Comments: TODO:
! 
! $Id: confusion.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module confusion

  use types

  implicit none 

  private 

  public findConfusedSets

contains 

  !---------------------------------------------------------------------
  subroutine findConfusedSets(srcX, srcY, srcRadius, confSetNum, confused)
  !---------------------------------------------------------------------
  ! The routine takes as inputs a list of source positions and returns a
  ! list of indices. An index of zero indicates that the source is not
  ! confused with any other in the set (ie, in no case does the test for
  ! confusion with any other source return TRUE). If the returned index
  ! for any source is greater than zero,
  ! that indicates that the source is confused with another having the same
  ! index. All sources of the same index (except where this value is zero,
  ! obviously) are members of the same confusion set or group. Such a group
  ! is defined by two conditions: (i) it must be possible to move from any
  ! member of the confusion group to any other purely by moving between
  ! members of pairs of confused sources, each of which shares the same
  ! index; (ii) it should not be possible to move in such fashion to any
  ! source _not_ sharing the group index. Note however that it is _not_
  ! necessarily true that, if i and j are members of the same confusion
  ! group, then i is confused with j.
  !
  ! There are no gaps in the returned sequence of confusion indices. Ie
  ! if i>1 is a valid confusion index, so is i-1.
  !---------------------------------------------------------------------

    implicit none

    real(double),   intent(in)  :: srcX(:), srcY(size(srcX))
    real(double),   intent(in)  :: srcRadius(size(srcX))
    integer(int16), intent(out) :: confSetNum(size(srcX))

    optional :: confused

    interface
      function confused(x, y, rad) result(areConfused)
        use types
        real(double), intent(in) :: x(2), y(2), rad(2)
        logical :: areConfused
      end function confused
    end interface

    integer(int16) :: numSrcs, i, j, k, maxNumConfusedSets, oldConfSetNum
    real(single) :: sumOfRadii, dX, dY, sepSquared
    real(double), allocatable :: tempSrcRadius(:)
    logical :: areConfused

    !-------------------------------------------------------------------
    ! Start

    confSetNum = 0         ! default values, indicating that no source
    maxNumConfusedSets = 0 ! is confused with another.

    numSrcs = size(srcX)
    allocate(tempSrcRadius(numSrcs))
    tempSrcRadius = max(srcRadius, 0.0d0) ! lifts any negative values to zero.

    do i = 1, numSrcs - 1
      do j = i + 1, numSrcs
        if (present(confused)) then
          areConfused = confused((/srcX(i), srcX(j)/), (/srcY(i), srcY(j)/)&
            , (/tempSrcRadius(i), tempSrcRadius(j)/))

        else
          areConfused = .false.
          sumOfRadii = tempSrcRadius(i) + tempSrcRadius(j)
          dX = srcX(i) - srcX(j)
          dY = srcY(i) - srcY(j)
          sepSquared = (dX * dX + dY * dY)
          if (sepSquared < sumOfRadii * sumOfRadii) areConfused = .true.
        end if

        if (areConfused) then

          if (confSetNum(i) > 0) then ! source i was previously found to
          ! be confused with at least one other source.

            if (confSetNum(j) > 0) then ! source j was previously found to
            ! be confused with at least one other source.

              if (confSetNum(i) .ne. confSetNum(j)) then ! sources i and j
              ! did not previously belong to the same confusion group.
              ! However since i and j have now been found to be confused,
              ! their groups must be merged into one. So, the numbers of
              ! the 'i' group are thus made the same as the numbers of
              ! the 'j' group:

                oldConfSetNum = confSetNum(i)
                do k = 1, numSrcs
                  if (confSetNum(k) == oldConfSetNum) then
                    confSetNum(k) = confSetNum(j)
                  end if
                end do

              else
              end if

            else ! source j is not confused with any other source. Add it
            ! to the confusion group of source i:

              confSetNum(j) = confSetNum(i)
            end if

          else ! source i is not confused with any other source.
            if (confSetNum(j) > 0) then ! source j was previously found to
            ! be confused with at least one other source. Add source i
            ! to the confusion group of source j:

              confSetNum(i) = confSetNum(j)

            else ! neither source i nor j is confused with another source.
            ! Start a new confusion group:

              maxNumConfusedSets = maxNumConfusedSets + 1
              confSetNum(j) = maxNumConfusedSets
              confSetNum(i) = maxNumConfusedSets
            end if
          end if
        end if
      end do
    end do

    ! Get rid of gaps in the sequence of confSetNum integers:
    do i = 1, maxNumConfusedSets
      if (.not. any(confSetNum == i)) then
        where(confSetNum > i) confSetNum = confSetNum - 1
      end if
    end do

  end subroutine findConfusedSets

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module confusion
