! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module:dss_aux
!
! Authors:  Ian Stewart - SSC/LUX
!        :  Masaaki Sakano - SSC/LUX
!          
! Description: Contains auxiliary routines to do with the data subspace (DSS).
! 
! Comments: TODO:
!	- change status values to error calls.
! 
! $Id: dss_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module dss_aux

  use dal
  use errorHandling
  use dss
  use caltypes	! for IntervalT

  use array_utils	! for smallestTrueIndex, largestTrueIndex
			! and trapezoidSum.
  use constants		! for lnTwo.
  use dal_aux
  use dss_ranges_aux	! for valueIsWithinRange
  use intervals_aux	! for andIntervals
  use sort_mod		! for quickSortStatic.
  use ssc_misc

  implicit none 

  private 

  public hasDss
  public dumpDss
  ! public dumpDssComponent
  public getNum1stDssCompWithValue
  private dssCompObeys
  public numDssCompsOfAxis
  public numDssCompsWithValue
  public valueIsPassedByDssComp
  public dssCompsIdentical
public dssCompsIdenticalList
  public dssCompNumsWhichPassThisValue

!!!  public determineUniqueFilterCombos

  public filterTypeToStr
  private filterOverlaps
  public anyFilterOfThisAxis
  public numFiltersOfAxis
  public get1stFilterThisAxis
  public allFiltersPassValue
  public valueIsPassedByFilter
  public filtersAreIdentical

  public getRangesSpan
  public readRanges
  public deallocateDssConstraint
  public deallocateDssCompSubset
  public deallocateDssCompSubsets
  private addNewCompSubset
  private addNewCompNumToSubset
  private copyCompSubsets
  private copyCompSubset
  private compSubsetsIdentical

!  public readGtis
  public readGtisFromDss
  public getGtiFromFilter
!  public andGtis

  public decomposeRegionXyNameString

  interface hasDss
    module procedure hasDssSet, hasDssArray, hasDssTab, hasDssBlock
  end interface

  interface readRanges
    module procedure readRangesDss
    module procedure readRangesComp
  end interface

  interface dssCompsIdentical
    module procedure dssCompsIdenticalInt
    module procedure dssCompsIdenticalComp
!    module procedure dssCompsIdenticalList, dssCompsIdenticalSimple
    module procedure dssCompsIdenticalSimple
    module procedure dssCompsIdentNumVecNameScal
    module procedure dssCompsIdentNumVecNameVec
  end interface

  interface dumpDss
    module procedure dumpDssBlock
    module procedure dumpDssPointer
    module procedure dumpDssComponent
  end interface

  interface readGtisFromDss
    module procedure readGtisFromDssPointer, readGtisFromDssBlock
  end interface

  interface deallocateDssCompSubsets
    module procedure deallocateDssCompSubsetOne, deallocateDssCompSubsetAry
  end interface

  integer, parameter :: &	! See dsslib/src/dsscore/dssinter.f90 for the definition of 80 and 255
       dssFiltCharLength = 80 &
     , dssFilterBlockCharLength = 255

  integer, parameter :: AryIdFilterT(7) &
     = (/ RANGEFILTERT, REGIONFILTERT, BITMASKFILTERT, RANGEBLOCKFILTERT, &
          GTIFILTERT, MASKFILTERT, REGIONBLOCKFILTERT /)
     ! see dsslib/src/dsscore/dssinter.f90 and/or the reference of Dsslib.
  integer, parameter :: SizeAryIdFilterT = size(AryIdFilterT)

  type, public :: dssConstraintType
    character(dssFiltCharLength)         :: name
    logical(bool)         :: ignore
    type(RangeT), pointer :: ranges(:) => null()
  end type dssConstraintType

  type, public :: dssCompSubsetType
    type(RangeT), pointer :: ranges(:) => null()
    integer, pointer :: compNums(:) => null()
  end type dssCompSubsetType

contains

  !---------------------------------------------------------------------
  logical(bool) function hasDssSet(set)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set

    !-------------------------------------------------------------------
    ! Start

    hasDssSet = .false. ! default
    if (hasAttribute(set, 'DSTYP1')) hasDssSet = .true.

  end function hasDssSet

  !---------------------------------------------------------------------
  logical(bool) function hasDssArray(inArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray

    !-------------------------------------------------------------------
    ! Start

    hasDssArray = .false. ! default
    if (hasAttribute(inArray, 'DSTYP1')) hasDssArray = .true.

  end function hasDssArray

  !---------------------------------------------------------------------
  logical(bool) function hasDssTab(tab)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab

    !-------------------------------------------------------------------
    ! Start

    hasDssTab = .false. ! default
    if (hasAttribute(tab, 'DSTYP1')) hasDssTab = .true.

  end function hasDssTab

  !---------------------------------------------------------------------
  logical(bool) function hasDssBlock(inBlock)
  !---------------------------------------------------------------------

    implicit none

    type(BlockT), intent(in) :: inBlock

    !-------------------------------------------------------------------
    ! Start

    hasDssBlock = .false. ! default
    if (hasAttribute(inBlock, 'DSTYP1')) hasDssBlock = .true.

  end function hasDssBlock

  !---------------------------------------------------------------------
  subroutine dumpDssBlock(inBlock)
  !---------------------------------------------------------------------

    implicit none

    type(BlockT), intent(in) :: inBlock

    type(DssT) :: dssPointer

    !-------------------------------------------------------------------
    ! Start:

    dssPointer = dataSubspace(inBlock)
    call dumpDssPointer(dssPointer)
    call dssRelease(dssPointer)

  end subroutine dumpDssBlock

  !---------------------------------------------------------------------
  subroutine dumpDssPointer(dssPointer)
  !---------------------------------------------------------------------

    implicit none

    type(DssT), intent(in) ::&
      dssPointer

    type(DScompT) :: dssComp
    integer :: numComps, compNum

    !-------------------------------------------------------------------
    ! Start:

    numComps = dssComponents(dssPointer)
    if (numComps <= 0) then
      print *, 'DSS contains no components.'
  return
    end if

    do compNum = 0, numComps - 1
      print *, 'DSS component number', compNum
      dssComp = dssComponent(dssPointer, compNum)
      call dumpDssComponent(dssComp)
    end do

  end subroutine dumpDssPointer

  !---------------------------------------------------------------------
  subroutine dumpDssComponent(dssComp)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT), intent(in) ::&
      dssComp

    type(DSfilterT) ::&
      filter
    type(RangeT), dimension(:), allocatable :: range 
    character(dssFilterBlockCharLength) ::&
      blockName
    integer(kind=int16) ::&
      i
!    integer(int32) ::&
!      bitmask
integer(int32), pointer ::&
  onBitMasks(:) => null(),&
  offBitMasks(:) => null()
    integer :: filtNum, filterType, nSize
    logical :: typeNotFound
    character(len=dssFiltCharLength) :: columnName, filterTypeStr
    character(len=dssFiltCharLength) :: axisNames
!    character(8) ::&
character(18) ::&
      bitMaskStr
    type filtInfo
      integer :: index, nSize
      character(len=dssFiltCharLength) :: name
    end type
    type(filtInfo) :: myFiltInfo(7)

    !-------------------------------------------------------------------
    ! Start:

    myFiltInfo(1)%index = RANGEFILTERT
    myFiltInfo(2)%index = REGIONFILTERT
    myFiltInfo(3)%index = BITMASKFILTERT
    myFiltInfo(4)%index = RANGEBLOCKFILTERT
    myFiltInfo(5)%index = GTIFILTERT
    myFiltInfo(6)%index = MASKFILTERT
    myFiltInfo(7)%index = REGIONBLOCKFILTERT
    myFiltInfo(1)%name = 'RANGEFILTERT'
    myFiltInfo(2)%name = 'REGIONFILTERT'
    myFiltInfo(3)%name = 'BITMASKFILTERT'
    myFiltInfo(4)%name = 'RANGEBLOCKFILTERT'
    myFiltInfo(5)%name = 'GTIFILTERT'
    myFiltInfo(6)%name = 'MASKFILTERT'
    myFiltInfo(7)%name = 'REGIONBLOCKFILTERT'

    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
      columnName = trim(dssFilterName(filter))
      filterType = dssFilterType(filter)

      typeNotFound = .true.
      do i = 1, 7
        if (myFiltInfo(i)%index .eq. filterType) then
          typeNotFound = .false.
          filterTypeStr = myFiltInfo(i)%name
      exit
        end if
      end do
      if (typeNotFound) then
        write(errStr, '(a, i3.3, a, i3.3, a)') "ssclib:dumpDssComponent: Filter "&
         , filtNum, " in the current DSS component has type integer ", filterType&
         , "; this is not recognized."
call error('unknownDssFilterType')
      end if

      write(*, '(a, i3.3, a)') '  Filter=', filtNum&
        , ', axis='//trim(columnName)//', type='//trim(filterTypeStr)

      select case(filterType)
      case(RANGEFILTERT)
      	call dssFilterRangesSize(filter,nSize)
      	allocate(range(0:nSize-1))
        range = dssRanges(filter)
        call dumpRanges(range, 4)
        deallocate(range)

      case(REGIONBLOCKFILTERT)
        blockName = trim(dssFilterBlock(filter))
        print *, '    Block name = ', trim(blockName)

      case(BITMASKFILTERT)
        call dssFilterMask(filter, onBitMasks, offBitMasks)
        print *, '    On-bit masks:'
        do i = lbound(onBitMasks, 1), ubound(onBitMasks, 1)
          write(bitMaskStr,  '(a5, i2, a3, z8)') 'mask ', i, ' = '&
            , onBitMasks(i)
          print *, '      ', bitMaskStr
        end do
        print *, '    Off-bit masks:'
        do i = lbound(offBitMasks, 1), ubound(offBitMasks, 1)
          write(bitMaskStr,  '(a5, i2, a3, z8)') 'mask ', i, ' = '&
            , offBitMasks(i)
          print *, '      ', bitMaskStr
        end do
!!!        deallocate(onBitMasks, offBitMasks)	! Should it be commented-out??

      case(RANGEBLOCKFILTERT)
        blockName = trim(dssFilterBlock(filter))
        print *, '    Block name = ', trim(blockName)

      case(GTIFILTERT)
        blockName = trim(dssFilterBlock(filter))
        print *, '    Block name = ', trim(blockName)

      case(MASKFILTERT, REGIONFILTERT)
        axisNames = trim(dssFilterComps(filter))
        print *, '    Axis names', trim(axisNames)
      end select
    end do
    print *, ''

  end subroutine dumpDssComponent

  !---------------------------------------------------------------------
  subroutine getNum1stDssCompWithValue(dssPointer, axisName, axisValue&
  , firstCompNum, filterType, trueIfNone, status)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue
    integer, intent(out) :: firstCompNum
    integer, intent(in), optional :: filterType	! (Optional) Single integer to specify the filterType of interest.  If not specified, all the filters are taken into account.
    logical, intent(in), optional :: trueIfNone	! (Optional) Passed to allFiltersPassValue().  If TRUE, the DssComp which has no (specified) filter is regarded as true (defaut and recommended).
    integer, intent(out), optional :: status	! (Optional) 0 in default.  If no DSS component found, it is -1 or raises an error if this argument is not given.

    integer :: compNum
    type(DScompT) :: dssComp

    character(dssFilterBlockCharLength) :: gtiBlockName
    integer         :: filtNum
    type(DSfilterT) :: filter
    character(dssFiltCharLength) :: tmpDssFiltName
logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0	! Default.

    firstCompNum = -1	! default to invalid value.

    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
if (testFlag) print *,'DEBUG: Checking allFiltersPassValue() with compNum=',compNum	!! for DEBUG
      ! if (allFiltersPassValue(dssComp, axisName, axisValue, filterType)) then
      if (allFiltersPassValue(dssComp, axisName, axisValue &
         , filterType=filterType, trueIfNone=trueIfNone)) then
        firstCompNum = compNum
if (testFlag) print *,'DEBUG: Passed allFiltersPassValue() with compNum=',compNum	!! for DEBUG
    exit
      end if
    end do

    ! Now evaluating the result.
    if (firstCompNum < 0) then
      write(errStr, '(a, es9.3, a)') "ssclib:getNum1stDssCompWithValue: &
        &No DSS component passes value ", axisValue, " on column "//trim(axisName)//"."
      if (present(status)) then
        status = -1	! Suggesting erroneous.
        call message(appMsg, Noisy)
      else
call error('noDssCompThisValue')
      end if
    else
      if (testFlag) then	!! for DEBUG
        gtiBlockName = ''
        do filtNum = 0, dssFilters(dssComp) - 1
          filter = dssFilter(dssComp, filtNum)
          tmpDssFiltName = dssFilterName(filter)
          if (trim(tmpDssFiltName) .eq. 'TIME') then
            gtiBlockName = dssFilterBlock(filter)
            exit
          end if
        end do
        write(errStr, '(a,i5,a,es9.3,a,i3,a,i3,a)') "ssclib:getNum1stDssCompWithValue: DSSComp=("&
           , firstCompNum, ') is returned on column '//trim(axisName)//"=(", axisValue &
           , "), where the corresponding block name is ("//trim(gtiBlockName) &
           //") and filter-No = (",filtNum,") out of (0 -- ",dssFilters(dssComp)-1,")."
        call message(appMsg, Noisy)
      end if
    end if

  end subroutine getNum1stDssCompWithValue

  !---------------------------------------------------------------------
  logical(bool) function dssCompObeys(dssComp, dssConstraints)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT),           intent(in) :: dssComp
    type(dssConstraintType), intent(in) :: dssConstraints(:)

    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: numConstraints, i, filtNum
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    dssCompObeys = .true. ! default

    numConstraints = size(dssConstraints)
    do i = 1, numConstraints
    if (numFiltersOfAxis(dssComp, dssConstraints(i)%name) .le. 0) cycle
      do filtNum = 0, dssFilters(dssComp) - 1
        filter = dssFilter(dssComp, filtNum)
        tmpDssFiltName = dssFilterName(filter)
      if (trim(tmpDssFiltName) .ne. trim(dssConstraints(i)%name)) cycle
      !if (trim(dssFilterName(filter)) .ne. dssConstraints(i)%name) cycle

        if (filterOverlaps(filter, dssConstraints(i))) then
          dssCompObeys = .false.
  return
        end if
      end do
    end do

  end function dssCompObeys

  !---------------------------------------------------------------------
  integer function numDssCompsOfAxis(dssPointer, axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName

    integer :: compNum
    type(DScompT) :: dssComp

    !-------------------------------------------------------------------
    ! Start

    numDssCompsOfAxis = 0
    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
      if (anyFilterOfThisAxis(dssComp, axisName))&
        numDssCompsOfAxis = numDssCompsOfAxis + 1
    end do

  end function numDssCompsOfAxis

  !---------------------------------------------------------------------
  integer function numDssCompsWithValue(dssPointer, axisName, axisValue, trueIfNone)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue
    logical,      intent(in), optional :: trueIfNone	! (Optional) Passed to allFiltersPassValue().  If TRUE, the DssComp which has no (specified) filter is regarded as true (defaut and recommended).

    integer :: compNum
    type(DScompT) :: dssComp
logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    numDssCompsWithValue = 0 ! default

    do compNum = 0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, compNum)
if (testFlag) print *,'DEBUG(numDssCompsWithValue): Checking allFiltersPassValue() with compNum=',compNum	!! for DEBUG
      ! if (allFiltersPassValue(dssComp, axisName, axisValue))&
      if (allFiltersPassValue(dssComp, axisName, axisValue, trueIfNone=trueIfNone))&
        numDssCompsWithValue = numDssCompsWithValue + 1
    end do

  end function numDssCompsWithValue

  !---------------------------------------------------------------------
  logical(bool) function valueIsPassedByDssComp(dssComp, axisName, axisValue)
  !---------------------------------------------------------------------
  ! This returns true ONLY if the specified component contains no filter on the specified axis OR the value is passed by all of the filters of this component that do exist for the given axis.

    implicit none

    type(DScompT),intent(in) :: dssComp
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue

    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    logical, allocatable :: filterIsOnSpecifiedAxis(:), valueIsPassed(:)

    !-------------------------------------------------------------------
    ! Start

!print *, '      aaa valueIsPassedByDssComp'

    numFilters = dssFilters(dssComp)
!print *, '      bbb'
    allocate(filterIsOnSpecifiedAxis(numFilters), valueIsPassed(numFilters))
!print *, '      ccc'
    filterIsOnSpecifiedAxis = .false.	! default
!print *, '      ddd'
    valueIsPassed           = .false.	! default

!print *, '      eee'
!print *, '      valueIsPassedByDssComp: numFilters=', numFilters
    do filtNum = 0, numFilters - 1
!print *, ''
!print *, '      filtNum:', filtNum
      filter = dssFilter(dssComp, filtNum)
!print *, '      fff'
      tmpDssFiltName = dssFilterName(filter)
      if (trim(tmpDssFiltName) .eq. trim(axisName)) then
      !if (trim(dssFilterName(filter)) .eq. axisName) then
!print *, '      ggg'
        filterIsOnSpecifiedAxis(filtNum) = .true.
!print *, '      hhh'
        if (valueIsPassedByFilter(axisValue, filter)) then
!print *, '      iii'
          valueIsPassed(filtNum) = .true.
        end if
      end if
    end do
!print *, '      jjj'

    if (all(filterIsOnSpecifiedAxis .eqv. valueIsPassed)) then
      valueIsPassedByDssComp = .true.
    else
      valueIsPassedByDssComp = .false.
    end if
!print *, '      zzz valueIsPassedByDssComp'

  end function valueIsPassedByDssComp

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdenticalInt(dssPointer, compNumA, compNumB&
  , axisNamesToInclude, axisNamesToNeglect)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in)           :: dssPointer
    integer,      intent(in)           :: compNumA, compNumB
    character(*), intent(in), optional :: axisNamesToInclude(:),&
                                          axisNamesToNeglect(:)

    type(DScompT) :: dssCompA, dssCompB

    !-------------------------------------------------------------------
    ! Start

    dssCompA = dssComponent(dssPointer, compNumA)
    dssCompB = dssComponent(dssPointer, compNumB)

!*** DON'T forget comp numbers start at ZERO

    if (present(axisNamesToInclude)) then
      if (present(axisNamesToNeglect)) then
        dssCompsIdenticalInt = dssCompsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToInclude, axisNamesToNeglect)

      else
        dssCompsIdenticalInt = dssCompsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToInclude)
      end if

    else
      if (present(axisNamesToNeglect)) then
        dssCompsIdenticalInt = dssCompsIdenticalComp(dssCompA , dssCompB&
          , axisNamesToNeglect=axisNamesToNeglect)

      else
        dssCompsIdenticalInt = dssCompsIdenticalComp(dssCompA , dssCompB)
      end if
    end if

  end function dssCompsIdenticalInt

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdentNumVecNameScal(dssPointer, dssCompNums&
  , axisName)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    integer,      intent(in) :: dssCompNums(:)
    character(*), intent(in) :: axisName

    !-------------------------------------------------------------------
    ! Start

    dssCompsIdentNumVecNameScal = dssCompsIdentNumVecNameVec(dssPointer&
      , dssCompNums, (/axisName/))

  end function dssCompsIdentNumVecNameScal

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdentNumVecNameVec(dssPointer, dssCompNums&
  , axisNames)
  !---------------------------------------------------------------------

    implicit none

    type(DssT),   intent(in) :: dssPointer
    integer,      intent(in) :: dssCompNums(:)
    character(*), intent(in) :: axisNames(:)

    integer :: numDssCompIndices, numAxes, compIndex, compNum, i
    type(DScompT) :: dssCompA, dssCompB

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa dssCompsIdentNumVecNameVec'

    numDssCompIndices = size(dssCompNums)
    numAxes           = size(axisNames)
if (testFlag) print *, 'bbb numDssCompIndices=',numDssCompIndices
if (testFlag) print *, 'ccc numAxes=',numAxes
    if (numDssCompIndices < 2) then
      dssCompsIdentNumVecNameVec = .true.
  return
    end if

    dssCompsIdentNumVecNameVec = .false. ! default

    compIndex = 1
    compNum = dssCompNums(compIndex)
    dssCompA = dssComponent(dssPointer, compNum)
    do compIndex = 2, numDssCompIndices
      compNum = dssCompNums(compIndex)
      dssCompB = dssComponent(dssPointer, compNum)
if (testFlag) print *, '  c0 2nd compIndex=',compIndex
if (testFlag) print *, '  c1 2nd compNum=',compNum

      do i = 1, numAxes
if (testFlag) print *, '    c1a axis num=',i
if (testFlag) print *, '    c1b axis name=',trim(axisNames(i))
        if (.not. dssCompsIdenticalSimple(dssCompA, dssCompB, axisNames(i))) then
if (testFlag) print *, 'zzz components NOT identical.'
  return
        else
if (testFlag) print *, '    c1c components are identical.'
        end if
      end do
    end do

    dssCompsIdentNumVecNameVec = .true. ! passed all its tests.
if (testFlag) print *, 'zzz dssCompsIdentNumVecNameVec'

  end function dssCompsIdentNumVecNameVec

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdenticalComp(dssCompA, dssCompB&
  , axisNamesToInclude, axisNamesToNeglect)
  !---------------------------------------------------------------------
!*** I don't think this works. Use List if you have no axisNamesToNeglect.


!**** NOTE comparison of reals is a bad idea - replace '==' by 'almost-equal'.

    implicit none

    type(DScompT), intent(in)           :: dssCompA, dssCompB
    character(*),  intent(in), optional :: axisNamesToInclude(:),&
                                           axisNamesToNeglect(:)

    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: numFiltersA, numFiltersB, filtNumA, filtNumB, i
    logical, allocatable :: compBFilterMatched(:)
    character(dssFiltCharLength) :: filterNameA, filterNameB
    logical :: unableToMatchFilter
    type(DSfilterT) :: filterA, filterB

    !-------------------------------------------------------------------
    ! Start

    dssCompsIdenticalComp = .false. ! default

    numFiltersA = dssFilters(dssCompA)
    numFiltersB = dssFilters(dssCompA)

    allocate(compBFilterMatched(numFiltersB))
    compBFilterMatched = .false. ! initially.
    DCICOUT: do filtNumA = 1, numFiltersA
      ! Skip over filterA if it either _isn't_ in the list of axis names
      ! to include, or it _is_ in the list of axis names to neglect.

      tmpDssFiltName = dssFilterName(dssFilter(dssCompA, filtNumA-1))
      filterNameA = trim(tmpDssFiltName)
      !filterNameA = trim(dssFilterName(dssFilter(dssCompA, filtNumA-1)))
      if (present(axisNamesToInclude)) then
    ! if (.not. any(filterNameA .eq. axisNamesToInclude)) cycle
        DCICIN1: do i = 1, size(axisNamesToInclude)
          if (filterNameA .ne. axisNamesToInclude(i)) cycle DCICOUT
        end do DCICIN1
      end if
      if (present(axisNamesToNeglect)) then
    ! if (any(filterNameA .eq. axisNamesToNeglect)) cycle
        DCICIN2: do i = 1, size(axisNamesToNeglect)
          if (filterNameA .eq. axisNamesToNeglect(i)) cycle DCICOUT
        end do DCICIN2
      end if

      unableToMatchFilter = .true. ! default
      DCICINMAIN: do filtNumB = 1, numFiltersB
      if (compBFilterMatched(filtNumB)) cycle

        tmpDssFiltName = dssFilterName(dssFilter(dssCompB, filtNumB-1))
        filterNameB = trim(tmpDssFiltName)
        !filterNameB = trim(dssFilterName(dssFilter(dssCompB, filtNumB-1)))
      if (filterNameA .ne. filterNameB) cycle

        ! Tick filterB off if it either _isn't_ in the list of axis names
        ! to include, or it _is_ in the list of axis names to neglect.

        if (present(axisNamesToInclude)) then
          ! if (.not. any(filterNameB .eq. axisNamesToInclude))&
          !   compBFilterMatched(filtNumB) = .true.
          DCICIN3: do i = 1, size(axisNamesToInclude)
            if (filterNameB .ne. axisNamesToInclude(i)) &
               compBFilterMatched(filtNumB) = .true.
          end do DCICIN3
        end if
        if (present(axisNamesToNeglect)) then
          ! if (any(filterNameB .eq. axisNamesToNeglect))&
          !   compBFilterMatched(filtNumB) = .true.
          DCICIN4: do i = 1, size(axisNamesToInclude)
            if (filterNameB .eq. axisNamesToNeglect(i)) &
               compBFilterMatched(filtNumB) = .true.
          end do DCICIN4
        end if

        filterA = dssFilter(dssCompA, filtNumA - 1)
        filterB = dssFilter(dssCompB, filtNumB - 1)
        if (filtersAreIdentical(filterA, filterB)) then
          unableToMatchFilter = .false.
          compBFilterMatched(filtNumB) = .true.
      exit
        end if
      end do DCICINMAIN

  if (unableToMatchFilter) return
    end do DCICOUT

  if (any(.not. compBFilterMatched)) return

    dssCompsIdenticalComp = .true. ! made it through all the tests.

  end function dssCompsIdenticalComp

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdenticalList(dssCompA, dssCompB, axisNames)
  !---------------------------------------------------------------------
  ! This returns true only if either (i) both components have no filter of the given axis name, or (ii) both components have the same number, type and values of filters of the given axis name.

    implicit none

    type(DScompT), intent(in) :: dssCompA, dssCompB
    character(*),  intent(in) :: axisNames(:)

    integer :: i

    !-------------------------------------------------------------------
    ! Start

    dssCompsIdenticalList = .false. ! default

    do i = 1, size(axisNames)
      if (.not. dssCompsIdenticalSimple(dssCompA, dssCompB, axisNames(i)))&
  return
    end do

    dssCompsIdenticalList = .true. ! passed all its tests.

  end function dssCompsIdenticalList

  !---------------------------------------------------------------------
  logical(bool) function dssCompsIdenticalSimple(dssCompA, dssCompB, axisName)
  !---------------------------------------------------------------------
  ! This returns true only if either (i) both components have no filter of the given axis name, or (ii) both components have the same number, type and values of filters of the given axis name.

    implicit none

    type(DScompT), intent(in) :: dssCompA, dssCompB
    character(*),  intent(in) :: axisName

    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: numFiltersA, numFiltersB, filtNumA, filtNumB, filtNumA1, filtNumA2, filtNumB1, filtNumB2
    logical, allocatable :: filtersThisAxisA(:), filtersThisAxisB(:)
    type(DSfilterT) :: filterA1, filterA2, filterB1, filterB2, filterA, filterB
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    dssCompsIdenticalSimple: axis name is '//trim(axisName)//'.'

    dssCompsIdenticalSimple = .false. ! default

    if   (.not. anyFilterOfThisAxis(dssCompA, axisName)&
    .and. .not. anyFilterOfThisAxis(dssCompB, axisName)) then
      dssCompsIdenticalSimple = .true.
if (testFlag) print *, '    dssCompsIdenticalSimple:   no filter this name in either comp.'
  return
    end if

    numFiltersA = dssFilters(dssCompA)
    allocate(filtersThisAxisA(numFiltersA))
    filtersThisAxisA = .false. ! default
    do filtNumA = 1, numFiltersA
      tmpDssFiltName = dssFilterName(dssFilter(dssCompA, filtNumA - 1))
      if (trim(tmpDssFiltName) == trim(axisName)) then
      !if (trim(dssFilterName(dssFilter(dssCompA, filtNumA - 1))) == axisName) then
!print *, 1
        filtersThisAxisA(filtNumA) = .true.
if (testFlag) print *, '    dssCompsIdenticalSimple:   compA has a filter of this name, filter number', filtNumA-1
end if
    end do

    ! Knock out duplicated dssCompA filters:

    do filtNumA1 = 1, numFiltersA - 1
    if (.not. filtersThisAxisA(filtNumA1)) cycle
      filterA1 = dssFilter(dssCompA, filtNumA1 - 1)
      do filtNumA2 = filtNumA1 + 1, numFiltersA
      if (.not. filtersThisAxisA(filtNumA2)) cycle
        filterA2 = dssFilter(dssCompA, filtNumA2 - 1)
        if (filtersAreIdentical(filterA1, filterA2))&
          filtersThisAxisA(filtNumA2) = .false.
      end do
    end do

    numFiltersB = dssFilters(dssCompB)
    allocate(filtersThisAxisB(numFiltersB))
    filtersThisAxisB = .false. ! default
    do filtNumB = 1, numFiltersB
      tmpDssFiltName = dssFilterName(dssFilter(dssCompB, filtNumB - 1))
      if (trim(tmpDssFiltName) == trim(axisName)) then
      !if (trim(dssFilterName(dssFilter(dssCompB, filtNumB - 1))) == axisName) then
!print *, 2
        filtersThisAxisB(filtNumB) = .true.
if (testFlag) print *, '    dssCompsIdenticalSimple:   compB has a filter of this name, filter number', filtNumB-1
end if
    end do

    ! Knock out duplicated dssCompB filters:

    do filtNumB1 = 1, numFiltersB - 1
    if (.not. filtersThisAxisB(filtNumB1)) cycle
      filterB1 = dssFilter(dssCompB, filtNumB1 - 1)
      do filtNumB2 = filtNumB1 + 1, numFiltersB
      if (.not. filtersThisAxisB(filtNumB2)) cycle
        filterB2 = dssFilter(dssCompB, filtNumB2 - 1)
        if (filtersAreIdentical(filterB1, filterB2))&
          filtersThisAxisB(filtNumB2) = .false.
      end do
    end do

    if (count(filtersThisAxisA) .ne. count(filtersThisAxisB)) then
!print *, 3
  return
    end if
!print *, 4

    ! Now see if filters from A and B all match up. Each match-up is marked by setting the appropriate element of filtersThisAxisA (an arbitary choice, could equally well have chosen B for the honour) to false. Thus if any elements of filtersThisAxisA remain true at the end of the looping, this means not every filter was matched.

    ALOOP: do filtNumA = 1, numFiltersA
    if (.not. filtersThisAxisA(filtNumA)) cycle
      filterA = dssFilter(dssCompA, filtNumA - 1)

      BLOOP: do filtNumB = 1, numFiltersB
      if (.not. filtersThisAxisB(filtNumB)) cycle
        filterB = dssFilter(dssCompB, filtNumB - 1)

if (testFlag) print *, '    dssCompsIdenticalSimple: test filter numbers '&
, filtNumA-1, ' and ', filtNumB-1 ,' for identity'
        if (filtersAreIdentical(filterA, filterB)) then
!print *, '5 - YES'
          filtersThisAxisA(filtNumA) = .false.
if (testFlag) print *, '    dssCompsIdenticalSimple: for filtNumA=',filtNumA-1&
,' and filtNumB=',filtNumB-1,', filters are identical. Exit Bloop.'
      exit BLOOP
        else
!print *, '5 - NO'
        end if

      end do BLOOP
    end do ALOOP

if (testFlag) print *, '    dssCompsIdenticalSimple: filters unmatched?', filtersThisAxisA

    if (.not. any(filtersThisAxisA)) dssCompsIdenticalSimple = .true.

if (testFlag) print *, '    dssCompsIdenticalSimple: comps are identical=', dssCompsIdenticalSimple

  end function dssCompsIdenticalSimple

  !---------------------------------------------------------------------
  subroutine dssCompNumsWhichPassThisValue(dssPointer, axisName, axisValue&
  , dssCompNums, trueIfNone)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate dssCompNums after use.

    implicit none

    type(DssT),   intent(in)  :: dssPointer
    character(*), intent(in)  :: axisName
    real(single), intent(in)  :: axisValue
    integer,      pointer     :: dssCompNums(:)
    logical,      intent(in), optional :: trueIfNone	! (Optional) Passed to allFiltersPassValue().  If TRUE, the DssComp which has no (specified) filter is regarded as true (defaut and recommended).

    integer :: numDssComps, compNum, numMatchingComps
    integer, allocatable :: tempCompNums(:)
    type(DScompT) :: dssComp
!logical, parameter :: testFlag = .true.
logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa dssCompNumsWhichPassThisValue'
if (testFlag) print *, '  bbb axisName, axisValue: ', trim(axisName), axisValue

    numDssComps = dssComponents(dssPointer)
if (testFlag) print *, '  ccc numDssComps=', numDssComps
    allocate(tempCompNums(numDssComps))
    numMatchingComps = 0
    do compNum = 0, numDssComps - 1
if (testFlag) print *, '  ddd compNum=', compNum
      dssComp = dssComponent(dssPointer, compNum)
      if (allFiltersPassValue(dssComp, axisName, axisValue, trueIfNone=trueIfNone)) then
if (testFlag) print *, '  eee allFiltersPassValue=T'
        numMatchingComps = numMatchingComps + 1
        tempCompNums(numMatchingComps) = compNum
      end if
    end do
if (testFlag) print *, '  fff numMatchingComps=', numMatchingComps
    allocate(dssCompNums(numMatchingComps))
    if (numMatchingComps > 0) dssCompNums = tempCompNums(1:numMatchingComps)

if (testFlag) print *, '  zzz dssCompNumsWhichPassThisValue'
  end subroutine dssCompNumsWhichPassThisValue

  !---------------------------------------------------------------------
  subroutine determineUniqueFilterCombos(dssPointer, dssCompNums, axisAName&
  , axisBName, uniqueABCombos)
  !---------------------------------------------------------------------
  ! Work through the list of components and find out how many unique sets of axisA-axisB filters there are for the present ccd.
  !
!*** test this with more complicated selections.

    implicit none

    type(DssT), intent(in) :: dssPointer
    integer, intent(in) :: dssCompNums(:)
    character(*), intent(in) :: axisAName, axisBName
    integer(int16), intent(out) :: uniqueABCombos(size(dssCompNums))

    integer(int16) :: numDssComps, numUniqueABFilters, startCompIndex, compIndex
    logical(bool), allocatable :: compIndexIsFree(:)
    integer :: startCompNum, dssCompNum
    type(DScompT) :: startComp, dssComp

    !-------------------------------------------------------------------
    ! Start

    numDssComps = size(dssCompNums)
  if (numDssComps <= 0) return

    allocate(compIndexIsFree(numDssComps))

    uniqueABCombos = 0		! default
    compIndexIsFree = .true.	! default
    numUniqueABFilters = 0	! Although there should be
    ! at least 1. The routine returns if not, a few lines earlier.

    do while(any(compIndexIsFree))
      numUniqueABFilters = numUniqueABFilters + 1
      startCompIndex = smallestTrueIndex(compIndexIsFree)
      ! smallestTrueIndex is in ssclib/array_utils.

      startCompNum = dssCompNums(startCompIndex)
      startComp = dssComponent(dssPointer, startCompNum)
      uniqueABCombos(startCompIndex) = numUniqueABFilters
      compIndexIsFree(startCompIndex) = .false.

      do compIndex = startCompIndex + 1, numDssComps
      if (.not. compIndexIsFree(compIndex)) cycle

        dssCompNum = dssCompNums(compIndex)
        dssComp = dssComponent(dssPointer, dssCompNum)

        if   (dssCompsIdentical(startComp, dssComp, axisAName)&
        .and. dssCompsIdentical(startComp, dssComp, axisBName)) then
          uniqueABCombos(compIndex) = numUniqueABFilters
          compIndexIsFree(compIndex) = .false.
        end if
      end do
    end do

    if (numUniqueABFilters == 0)&
call error('dss_auxBug0',"ssclib:determineUniqueFilterCombos: please contact the code developer.")
    if (numUniqueABFilters > numDssComps)&
call error('dss_auxBug1',"ssclib:determineUniqueFilterCombos: please contact the code developer.")
    if (any(uniqueABCombos <= 0))&
call error('dss_auxBug2',"ssclib:determineUniqueFilterCombos: please contact the code developer.")

  end subroutine determineUniqueFilterCombos

  !---------------------------------------------------------------------
  function filterTypeToStr(filterType) result(filterTypeStr)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: filterType
    character(dssFiltCharLength) :: filterTypeStr

    !-------------------------------------------------------------------
    ! Start

    select case(filterType)
    case(RANGEFILTERT)
      filterTypeStr = 'RANGEFILTERT'
    case(REGIONFILTERT)
      filterTypeStr = 'REGIONFILTERT'
    case(BITMASKFILTERT)
      filterTypeStr = 'BITMASKFILTERT'
    case(RANGEBLOCKFILTERT)
      filterTypeStr = 'RANGEBLOCKFILTERT'
    case(MASKFILTERT)
      filterTypeStr = 'MASKFILTERT'
    case(REGIONBLOCKFILTERT)
      filterTypeStr = 'REGIONBLOCKFILTERT'
    case(GTIFILTERT)
      filterTypeStr = 'GTIFILTERT'
    case default
      filterTypeStr = 'UNKNOWN'
    end select

  end function filterTypeToStr

  !---------------------------------------------------------------------
  logical(bool) function filterOverlaps(filter, dssConstraint)
  !---------------------------------------------------------------------

    implicit none

    type(DSfilterT),         intent(in) :: filter
    type(dssConstraintType), intent(in) :: dssConstraint

    type(RangeT), allocatable :: ranges(:) 
    type(RangeT), pointer :: newRanges(:) => null()
	integer 	:: nSize
    !-------------------------------------------------------------------
    ! Start

    filterOverlaps = .true. ! default

  if (dssFilterType(filter) .ne. RANGEFILTERT) return
  if (size(dssConstraint%ranges) <= 0) return
	
	call dssFilterRangesSize(filter,nSize)
   	allocate(ranges(0:nSize-1))
    ranges = dssRanges(filter)
    if (size(ranges) <= 0) then
      deallocate(ranges)
  return
    end if

!!    newRanges = andRangesPair(oldToNew(ranges), dssConstraint%ranges)
!    newRanges => andRangesPair(ranges, dssConstraint%ranges)
call andRanges(ranges, dssConstraint%ranges, newRanges)
    if (size(newRanges) <= 0) filterOverlaps = .false.

    deallocate(ranges, newRanges)

  end function filterOverlaps

  !---------------------------------------------------------------------
  logical(bool) function anyFilterOfThisAxis(dssComp, axisName)
  !---------------------------------------------------------------------
  ! This returns true only if the specified component contains at least one filter on the specified axis.

    implicit none

    type(DScompT),   intent(in) :: dssComp
    character(*), intent(in) :: axisName

    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: filtNum, numFilters
    type(DSfilterT) :: filter

    !-------------------------------------------------------------------
    ! Start

    numFilters = dssFilters(dssComp)
    anyFilterOfThisAxis = .false.	! default

    do filtNum = 0, numFilters - 1
      filter = dssFilter(dssComp, filtNum)
      tmpDssFiltName = dssFilterName(filter)	! Two lines can get together?
      if (trim(tmpDssFiltName) .eq. trim(axisName)) then
      !if (trim(dssFilterName(filter)) .eq. axisName) then
        anyFilterOfThisAxis = .true.
  return
      end if
    end do

  end function anyFilterOfThisAxis

  !---------------------------------------------------------------------
  function numFiltersOfAxis(dssComp, axisName) result(numFiltAxis)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT), intent(in) :: dssComp
    character(*),  intent(in) :: axisName
!    real(single),  intent(in) :: axisValue
    integer :: numFiltAxis

    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    character(dssFiltCharLength) :: dssFilterNameChr

    !-------------------------------------------------------------------
    ! Start

!print *, '    aaa numFiltersOfAxis'
    numFiltAxis = 0  ! default
!print *, '    bbb'

    numFilters = dssFilters(dssComp)
!print *, '    ccc numFilters = ', numFilters 
    do filtNum = 0, numFilters - 1
!print *, '    ddd = ', filtNum
      filter = dssFilter(dssComp, filtNum)
!print *, '    eee'
!print *, '    MASfff'
      dssFilterNameChr = ''
      dssFilterNameChr = trim(dssFilterName(filter))
!print *, '    MASggg = ',trim(dssFilterNameChr)
      !if (trim(dssFilterNameChr) .eq. trim(axisName))&
      !if (trim(dssFilterName(filter)) .eq. trim(axisName))&	! SegFault
      !if (trim(dssFilterName(filter)) .eq. axisName)&
      if (dssFilterNameChr .eq. axisName)&
        numFiltAxis = numFiltAxis + 1 
!print *, '    MAShhh'
    end do
!print *, '    zzz numFiltersOfAxis'

  end function numFiltersOfAxis

  !---------------------------------------------------------------------
  subroutine get1stFilterThisAxis(dssComp, axisName, firstFilter)
  !---------------------------------------------------------------------

    implicit none

    type(DScompT),   intent(in)  :: dssComp
    character(*),    intent(in)  :: axisName
    type(DSfilterT), intent(out) :: firstFilter

    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    logical(bool) :: noFilterThisAxis

    !-------------------------------------------------------------------
    ! Start

    noFilterThisAxis = .true. ! default to bad status value.

    numFilters = dssFilters(dssComp)
    do filtNum = 0, numFilters - 1
      filter = dssFilter(dssComp, filtNum)
      tmpDssFiltName = dssFilterName(filter)
      if (trim(tmpDssFiltName) .eq. trim(axisName)) then
      !if (trim(dssFilterName(filter)) .eq. axisName) then
        noFilterThisAxis = .false.
        firstFilter = filter
    exit
      end if
    end do

    if (noFilterThisAxis)&
call error('noDssFilterThisAxis',"ssclib:get1stFilterThisAxis: The current DSS &
&component contains no filter on column "//trim(axisName)//".")

  end subroutine get1stFilterThisAxis

  !---------------------------------------------------------------------
  logical(bool) function allFiltersPassValue(dssComp, axisName, axisValue &
     , filterType, trueIfNone) result(isTrue)
  !---------------------------------------------------------------------
  ! This returns true EITHER (1) if the specified component contains no filter
  ! on the specified axis AND trueIfNone==.true. (default), or
  ! (2) all of the filters of this component and on this axis pass the specified value.
  !
  ! DSSLib reference states:
  !  "It is assumed that if a particular axis does
  !   not have a filter specied, then all values for that axis are allowed."
  ! For that reason, trueIfNone is TRUE in default.
  !
  ! NOTE: In the original version, it calls dssFilterName(filter),
  !  which may cause Segmenation Fault.
  !  Therefore it is rewritten with dssHasFilter().
  !  The algorithm is not completely identical, however, in the practical cases
  !  it should not cause any trouble, especially if the FITS files
  !  do not contain multiple DSS in the same dssComp with the same 
  !  axis name and filterName (which should not be the case in any FITS file).
  !  Note that the original and revised algorithms are switched via a parameter
  !  isUsed_dssFilterName .
  !
  ! ALGORITHM (the revised one): 
  !  (0)Prepare the array AryIdFilterT=(/ RANGEFILTERT, REGIONFILTERT, ... /).
  !  Then, the loop over i=(1..7):
  !  (1)Check whether a filter with the given axisName for AryIdFilterT(i) exists.
  !  (2)If so, check the consistency with the given axisValue in the filter.
  !  (3)Those results are stored in 
  !     filterOnThisAxisExists(i) and valueIsPassed(i)
  !     For example, if i==2, then that is for filterType==REGIONFILTERT (for the given dssComp and axisName).
  !  Finally returns true/false, as described at the top of this comments here.

    implicit none

    type(DScompT),   intent(in) :: dssComp
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue
    integer, intent(in), optional :: filterType	! (Optional) Single integer to specify the filterType of interest, if any.
    logical, intent(in), optional :: trueIfNone

    integer :: iFiltType
    type(DSfilterT) :: filter
    logical, allocatable :: filterOnThisAxisExists(:), valueIsPassed(:)
    ! logical, dimension(SizeAryIdFilterT) :: filterOnThisAxisExists, valueIsPassed	! SizeAryIdFilterT defined at the top of this code.

    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: filtNum, numFilters
    logical :: localTrueIfNone

logical, parameter :: isUsed_dssFilterName = .false.
logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    aaa allFiltersPassValue -- ', isUsed_dssFilterName
if (testFlag) print *, '    bbb axisName, axisValue: ', trim(axisName), axisValue

    if (present(trueIfNone)) then
      localTrueIfNone = trueIfNone
    else
      localTrueIfNone = .true.
    end if

    if (.not. isUsed_dssFilterName) then

    allocate(filterOnThisAxisExists(SizeAryIdFilterT), valueIsPassed(SizeAryIdFilterT))	! SizeAryIdFilterT defined at the top of this code.
    filterOnThisAxisExists = .false.	! default
    valueIsPassed          = .false.	! default

    do iFiltType = 1, SizeAryIdFilterT
      if (present(filterType)) then
        if (AryIdFilterT(iFiltType) .ne. filterType) cycle
      end if
      if (dssHasFilter(dssComp, trim(axisName), AryIdFilterT(iFiltType))) then	! AryIdFilterT defined at the top of this code.
        filter = dssFilter(dssComp, trim(axisName), AryIdFilterT(iFiltType))
        filterOnThisAxisExists(iFiltType) = .true.
        if (valueIsPassedByFilter(axisValue, filter))&
          valueIsPassed(iFiltType) = .true. 
      end if
if (testFlag) print *, '    eee filterOnThisAxisExists=', filterOnThisAxisExists(iFiltType)
if (testFlag) print *, '    fff valueIsPassed=', valueIsPassed(iFiltType)
    end do


    else	! if (.not. isUsed_dssFilterName) then
    !! Below is the original code, with the use of dssFilterName().

    numFilters = dssFilters(dssComp)
if (testFlag) print *, '    ccc numFilters=', numFilters
    allocate(filterOnThisAxisExists(0:numFilters - 1)&
    , valueIsPassed(0:numFilters - 1))
    filterOnThisAxisExists = .false.	! default
    valueIsPassed          = .false.	! default

    do filtNum = 0, numFilters - 1
if (testFlag) print *, '    ddd filtNum=', filtNum
      filter = dssFilter(dssComp, filtNum)
      tmpDssFiltName = dssFilterName(filter)
      if (trim(tmpDssFiltName) .eq. trim(axisName)) then
      !if (trim(dssFilterName(filter)) .eq. axisName) then
        filterOnThisAxisExists(filtNum) = .true.
        if (valueIsPassedByFilter(axisValue, filter))&
          valueIsPassed(filtNum) = .true. 
      end if
if (testFlag) print *, '    eee filterOnThisAxisExists=', filterOnThisAxisExists(filtNum)
if (testFlag) print *, '    fff valueIsPassed=', valueIsPassed(filtNum)
    end do

    end if	! if (.not. isUsed_dssFilterName) then

   ! if (.not. any(filterOnThisAxisExists) .or. any(filterOnThisAxisExists&
   ! .and. .not. valueIsPassed)) then	! This would return F if (! filterOnThisAxisExists(:)).
    if (.not. any(filterOnThisAxisExists)) then
      write(errStr, '(a,l1)') "ssclib:allFiltersPassValue: &
         &No (specified) filter exists for the axis ("//trim(axisName) &
         //'), hence it returns ', localTrueIfNone
      call message(appMsg, Noisy)
      isTrue = localTrueIfNone
    else if (any(filterOnThisAxisExists .and. .not. valueIsPassed)) then
      isTrue = .false.
    else
      isTrue = .true.
    end if
! print *,'DEBUG: filterOnThisA:', filterOnThisAxisExists
! print *,'DEBUG: valueIsPassed:', valueIsPassed, ', returning:', isTrue
! ! print *,'DEBUG: returning:', isTrue
if (testFlag) print *, '    ggg allFiltersPassValue=', isTrue

if (testFlag) print *, '    zzz allFiltersPassValue'

  end function allFiltersPassValue


  ! NOTE: In this original it calls dssFilterName(filter),
  !  which may cause a Segmentation Fault.
  !  Therefore the work-around is written.
  !
  !---------------------------------------------------------------------
  logical(bool) function allFiltersPassValue_org(dssComp, axisName, axisValue) result(isTrue)
  !---------------------------------------------------------------------
  ! This returns true EITHER if the specified component contains no filter on the specified axis OR all of the filters of this component and on this axis pass the specified value.

    implicit none

    type(DScompT),   intent(in) :: dssComp
    character(*), intent(in) :: axisName
    real(single), intent(in) :: axisValue

    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: filtNum, numFilters
    type(DSfilterT) :: filter
    logical, allocatable :: filterOnThisAxisExists(:), valueIsPassed(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '    aaa allFiltersPassValue'
if (testFlag) print *, '    bbb axisName, axisValue: ', trim(axisName), axisValue

    numFilters = dssFilters(dssComp)
if (testFlag) print *, '    ccc numFilters=', numFilters
    allocate(filterOnThisAxisExists(0:numFilters - 1)&
    , valueIsPassed(0:numFilters - 1))
    filterOnThisAxisExists = .false.	! default
    valueIsPassed          = .false.	! default

    do filtNum = 0, numFilters - 1
if (testFlag) print *, '    ddd filtNum=', filtNum
      filter = dssFilter(dssComp, filtNum)
      tmpDssFiltName = dssFilterName(filter)
      if (trim(tmpDssFiltName) .eq. trim(axisName)) then
      !if (trim(dssFilterName(filter)) .eq. axisName) then
        filterOnThisAxisExists(filtNum) = .true.
        if (valueIsPassedByFilter(axisValue, filter))&
          valueIsPassed(filtNum) = .true. 
      end if
if (testFlag) print *, '    eee filterOnThisAxisExists=', filterOnThisAxisExists(filtNum)
if (testFlag) print *, '    fff valueIsPassed=', valueIsPassed(filtNum)
    end do

!    if (.not. any(filterOnThisAxisExists) .or. any(filterOnThisAxisExists&
!    .and. .not. valueIsPassed)) then
if (any(filterOnThisAxisExists .and. .not. valueIsPassed)) then
      ! allFiltersPassValue = .false.
      isTrue = .false.
    else
      ! allFiltersPassValue = .true.
      isTrue = .true.
    end if
!if (testFlag) print *, '    ggg allFiltersPassValue=', allFiltersPassValue
if (testFlag) print *, '    ggg allFiltersPassValue=', isTrue

if (testFlag) print *, '    zzz allFiltersPassValue'

  end function allFiltersPassValue_org

  !---------------------------------------------------------------------
  logical(bool) function valueIsPassedByFilter(value, filter)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: value
    type(DSfilterT), intent(in) :: filter

    type(RangeT), allocatable :: ranges(:) 
    character(dssFilterBlockCharLength)  :: gtiBlockName
    type(DataSetT)  :: set
    type(TableT)    :: tab
    integer         :: numGtis, i, filterType, nSize
    type(RangeT), allocatable :: gti(:)
    logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    valueIsPassedByFilter = .false.

if (testFlag) print *,'DEBUG:ssclib:valueIsPassedByFilter(): Starting...'
!    select case(dssFilterType(filter))
filterType = dssFilterType(filter)
    select case(filterType)
    case(RANGEFILTERT)
if (testFlag) print *,' DEBUG:ssclib:valueIsPassedByFilter(): Processing RANGEFILTERT'
      call dssFilterRangesSize(filter,nSize)
      allocate(ranges(0:nSize-1))
      ranges = dssRanges(filter)
      do i = 0, size(ranges) - 1
        if (valueIsWithinRange(value, ranges(i))) then
          valueIsPassedByFilter = .true.
      exit
        end if
      end do

      deallocate(ranges)

!    case(REGIONFILTERT)
!      selectionExpression = selectExpression(filter)
!*** for a future version of valueIsPassedByFilter() with 2-value
!*** alternative input: set up pseudo-table with 1 entry, filter it,
!*** count num filtered rows.
!    case(BITMASKFILTERT)
!    case(RANGEBLOCKFILTERT)
!    case(MASKFILTERT)
!    case(REGIONBLOCKFILTERT)

    case(GTIFILTERT) ! Treat GTIs as ranges:
if (testFlag) print *,' DEBUG:ssclib:valueIsPassedByFilter(): Processing GTIFILTERT'
      gtiBlockName = dssFilterBlock(filter)
      ! gtiBlockName = trim(dssFilterBlock(filter))
if (testFlag) print *,' DEBUG:ssclib:valueIsPassedByFilter(): Reading '//trim(gtiBlockName) &
//':'//trim(gtiBlockName)
      set = dataSet(trim(gtiBlockName), READ)	! Is this really correct??!!!
      tab = table(set, trim(gtiBlockName))
      !tab = table(dataSet(gtiBlockName, READ), gtiBlockName)
      numGtis = numberOfRows(tab)
      allocate(gti(numGtis))
      call readColDataToFixed(tab, 'START', gti%lower%value)
      call readColDataToFixed(tab, 'STOP',  gti%upper%value)
      call release(tab)
      call release(set)

      gti%lower%type = INCLUSIVE
      gti%upper%type = INCLUSIVE
      do i = 1, size(gti)
        if (valueIsWithinRange(value, gti(i))) then
          valueIsPassedByFilter = .true.
      exit
        end if
      end do

    case default
      write(errStr, '(a, i3.3, a, i3.3, a, i3.3, a)') "ssclib:valueIsPassedByFilter: &
      &The filter of the current DSS component is of type ", filterType&
      , "; however it must be either a range filter (type ", RANGEFILTERT&
      , ") or a GTI filter (type ", GTIFILTERT, ")."
call error('badDssFilterType')
    end select

if (testFlag) print *,'DEBUG:ssclib:valueIsPassedByFilter(): Ends, returning ',valueIsPassedByFilter

  end function valueIsPassedByFilter

  !---------------------------------------------------------------------
  logical(bool) function filtersAreIdentical(filterA, filterB)
  !---------------------------------------------------------------------

!**** NOTE comparison of reals is a bad idea - replace '==' by 'almost-equal'.

    implicit none

    type(DSfilterT), intent(in) :: filterA, filterB

    type(RangeT), pointer :: rangesA(:) => null(), rangesB(:) => null()
    integer :: i, numRowsA, numRowsB, row, filterType
    character(dssFilterBlockCharLength) :: blockNameA, blockNameB
    character(dssFiltCharLength) :: xyNameString, xColName, yColName, shapeA, shapeB
    type(DataSetT) :: setA, setB
    type(TableT) :: tableA, tableB
    real(double), pointer :: colPtrA(:,:) => null(), colPtrB(:,:) => null()
    type(ColumnT) :: colA, colB
    type(ArrayT) :: arrayA, arrayB
!    integer(int32), pointer :: imageA(:,:) => null(), imageB(:,:) => null()
!logical(bool), pointer :: imageA(:,:) => null(), imageB(:,:) => null()
logical, pointer :: imageA(:,:) => null(), imageB(:,:) => null()
integer(int32), pointer ::&
  onBitMasksA(:) => null(), onBitMasksB(:) => null()&
 ,offBitMasksA(:) => null(), offBitMasksB(:) => null()

logical :: testFlag = .false.
character(dssFiltCharLength) :: strA, strB
integer :: intA, intB, nSize

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa filtersAreIdentical'

    filtersAreIdentical = .false. ! default

!!!  if (dssFilterName(filterA)  .ne. dssFilterName(filterB))  return
strA = dssFilterName(filterA)
strB = dssFilterName(filterB)
if (trim(strA) .ne. trim(strB))  return

if (testFlag) print *, '  1'

!!!  if (dssFilterType(filterA)  .ne. dssFilterType(filterB))  return
intA = dssFilterType(filterA)
intB = dssFilterType(filterB)
if (intA .ne. intB)  return

if (testFlag) print *, '  2'

!!!  if (dssFilterUnits(filterA) .ne. dssFilterUnits(filterB)) return
strA = dssFilterUnits(filterA)
strB = dssFilterUnits(filterB)
if (testFlag) print *, '*'//trim(strA)//'*'
if (testFlag) print *, '$'//trim(strB)//'$'
if (trim(strA) .ne. trim(strB)) return

if (testFlag) print *, '  3'

!    select case(dssFilterType(filterA))
    filterType = dssFilterType(filterA)
    select case(filterType)
    case(RANGEFILTERT)
if (testFlag) print *, '  RANGEFILTERT'
!      rangesA => oldToNewPtr(dssRanges(filterA))
!      rangesB => oldToNewPtr(dssRanges(filterB))
call dssFilterRangesSize(filterA,nSize)
allocate(rangesA(0:nSize-1))
rangesA = dssRanges(filterA)
call dssFilterRangesSize(filterB,nSize)
allocate(rangesB(0:nSize-1))
rangesB = dssRanges(filterB)

  if (size(rangesA) .ne. size(rangesB)) then
    deallocate(rangesA, rangesB)
    return
  end if

      do i = 0, size(rangesA) - 1
        if  (rangesA(i)%lower%type  .ne. rangesB(i)%lower%type&
        .or. rangesA(i)%lower%value .ne. rangesB(i)%lower%value&
        .or. rangesA(i)%upper%type  .ne. rangesB(i)%upper%type&
        .or. rangesA(i)%upper%value .ne. rangesB(i)%upper%value) then
          deallocate(rangesA, rangesB)
          return
        end if
      end do

    case(REGIONBLOCKFILTERT)
      blockNameA = trim(dssFilterBlock(filterA))
      blockNameB = trim(dssFilterBlock(filterB))
      setA = dataSet(blockNameA, READ)
      setB = dataSet(blockNameB, READ)
      tableA = table(setA, blockNameA)
      tableB = table(setB, blockNameB)
      !tableA = table(dataSet(blockNameA, READ), blockNameA)
      !tableB = table(dataSet(blockNameB, READ), blockNameB)
      if (.not. columnSetIsIdentical(tableA, tableB)) then ! in dal_aux
        call release(tableB)
        call release(tableA)
        call release(setB)
        call release(setA)
  return
      end if

      numRowsA = numberOfRows(tableA)
      numRowsB = numberOfRows(tableB)
      if (numRowsA .ne. numRowsB) then
        call release(tableB)
        call release(tableA)
        call release(setB)
        call release(setA)
  return
      end if

      xyNameString = stringAttribute(tableA, 'MFORM1')
      call decomposeRegionXyNameString(xyNameString, xColName, yColName)

      call readColDataToPtr(tableA, xColName, colPtrA)
      call readColDataToPtr(tableB, xColName, colPtrB)
!***** these readColDatas might be better toFixed, then wouldn't have to keep deallocating the darn pointers!

      if (any(colPtrA .ne. colPtrB)) then
        call release(tableB)
        call release(tableA)
        deallocate(colPtrA, colPtrB)
        call release(setB)
        call release(setA)
  return
      end if
!*** possibly could be fooled by nulls

      deallocate(colPtrA, colPtrB)

      call readColDataToPtr(tableA, yColName, colPtrA)
      call readColDataToPtr(tableB, yColName, colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableB)
        call release(tableA)
        deallocate(colPtrA, colPtrB)
        call release(setB)
        call release(setA)
  return
      end if
!*** possibly could be fooled by nulls

      deallocate(colPtrA, colPtrB)

      if(hasColumn(tableA, 'SHAPE')) then
        colA = column(tableA, 'SHAPE')
        colB = column(tableB, 'SHAPE')
        do row = 0, numRowsA - 1
          shapeA = ucase(stringCell(colA, row))
          shapeB = ucase(stringCell(colB, row))
          if (shapeA .ne. shapeB) then
            call release(colA)
            call release(colB)
            call release(tableA)
            call release(tableB)
            call release(setB)
            call release(setA)
  return
          end if
        end do
        call release(colA)
        call release(colB)
      end if

      if(hasColumn(tableA, 'R')) then
        call readColDataToPtr(tableA, 'R', colPtrA)
        call readColDataToPtr(tableB, 'R', colPtrB)
        if (any(colPtrA .ne. colPtrB)) then
          call release(tableB)
          call release(tableA)
          deallocate(colPtrA, colPtrB)
          call release(setB)
          call release(setA)
  return
        end if
!*** possibly could be fooled by nulls
      end if

      deallocate(colPtrA, colPtrB)

      if(hasColumn(tableA, 'ROTANG')) then
        call readColDataToPtr(tableA, 'ROTANG', colPtrA)
        call readColDataToPtr(tableB, 'ROTANG', colPtrB)
        if (any(colPtrA .ne. colPtrB)) then
          call release(tableB)
          call release(tableA)
          deallocate(colPtrA, colPtrB)
          call release(setB)
          call release(setA)
  return
        end if
!*** possibly could be fooled by nulls
      end if

      call release(tableB)
      call release(tableA)
      deallocate(colPtrA, colPtrB)
      call release(setB)
      call release(setA)

    case(BITMASKFILTERT)
!  if(dssFilterMask(filterA) .ne. dssFilterMask(filterB)) return

!*** for devtrack replace this by the following:
call dssFilterMask(filterA, onBitMasksA, offBitMasksA)
call dssFilterMask(filterB, onBitMasksB, offBitMasksB)
if (size(onBitMasksA) .ne. size(onBitMasksB)) then
  deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)
return
end if
if (size(offBitMasksA) .ne. size(offBitMasksB)) then
  deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)
return
end if
if (any(onBitMasksA .ne. onBitMasksB)) then
  deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)
return
end if
if (any(offBitMasksA .ne. offBitMasksB)) then
  deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)
return
end if

deallocate(onBitMasksA, offBitMasksA, onBitMasksB, offBitMasksB)

    case(GTIFILTERT)
      blockNameA = trim(dssFilterBlock(filterA))
      blockNameB = trim(dssFilterBlock(filterB))
      setA = dataSet(blockNameA, READ)
      setB = dataSet(blockNameB, READ)
      tableA = table(setA, blockNameA)
      tableB = table(setB, blockNameB)
      !tableA = table(dataSet(blockNameA, READ), blockNameA)
      !tableB = table(dataSet(blockNameB, READ), blockNameB)

      numRowsA = numberOfRows(tableA)
      numRowsB = numberOfRows(tableB)
      if (numRowsA .ne. numRowsB) then
        call release(tableB)
        call release(tableA)
        call release(setB)
        call release(setA)
  return
      end if

      call readColDataToPtr(tableA, 'TSTART', colPtrA)
      call readColDataToPtr(tableB, 'TSTART', colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableB)
        call release(tableA)
        deallocate(colPtrA, colPtrB)
        call release(setB)
        call release(setA)
  return
      end if
!*** possibly could be fooled by nulls

      deallocate(colPtrA, colPtrB)

      call readColDataToPtr(tableA, 'TSTOP', colPtrA)
      call readColDataToPtr(tableB, 'TSTOP', colPtrB)
      if (any(colPtrA .ne. colPtrB)) then
        call release(tableB)
        call release(tableA)
        deallocate(colPtrA, colPtrB)
        call release(setB)
        call release(setA)
  return
      end if
!*** possibly could be fooled by nulls

      call release(tableB)
      call release(tableA)
      deallocate(colPtrA, colPtrB)
      call release(setB)
      call release(setA)

    case(MASKFILTERT)
      blockNameA = trim(dssFilterBlock(filterA))
      blockNameB = trim(dssFilterBlock(filterB))
      setA = dataSet(blockNameA, READ)
      setB = dataSet(blockNameB, READ)
      arrayA = array(setA, blockNameA, READ)
      arrayB = array(setB, blockNameB, READ)
      !arrayA = array(dataSet(blockNameA, READ), blockNameA, READ)
      !arrayB = array(dataSet(blockNameB, READ), blockNameB, READ)
      call readArrayData(arrayA, imageA)
      call readArrayData(arrayB, imageB)

      if  (size(imageA, 1) .ne. size(imageB, 1)&
      .or. size(imageA, 2) .ne. size(imageB, 2)) then
        deallocate(imageA, imageB)
        call release(arrayB)
        call release(arrayA)
        call release(setB)
        call release(setA)
  return
      end if

!      if  (any(imageA .ne. imageB)) then
if  (any(imageA .neqv. imageB)) then
        deallocate(imageA, imageB)
        call release(arrayB)
        call release(arrayA)
        call release(setB)
        call release(setA)
  return
      end if

      deallocate(imageA, imageB)
      call release(arrayB)
      call release(arrayA)
      call release(setB)
      call release(setA)

!    case(RANGEBLOCKFILTERT)	! doesn't seem to be any firm convention
!				! for it yet.

!    case(REGIONFILTERT)	! not normally used in the DSS.

    case default
      write(errStr, '(a, i3.3, a)') "ssclib:filtersAreIdentical: The input &
        &filters are of type ", filterType, ", which is not recognized."
call error('badDssFilterType')
    end select

    filtersAreIdentical = .true. ! Made it through all the tests!
if (testFlag) print *, 'zzz filtersAreIdentical'

  end function filtersAreIdentical

  !---------------------------------------------------------------------
  function getRangesSpan(dssPointer, axisName) result(dssRangeSpan)
  !---------------------------------------------------------------------
  ! This subroutine does a logical OR of all the ranges of all the
  ! filters on axisName in all the components in the DSS and returns
  ! the extrema of that ORed range.

    implicit none

    type(DssT),   intent(in) :: dssPointer
    character(*), intent(in) :: axisName
    type(RangeT) :: dssRangeSpan

    type(DScompT) :: dssComp
    type(RangeT), pointer  :: oredRanges(:) => null(), compRanges(:) => null()
!, tempOredRanges(:)
    integer(int32) :: numComps, compNum, numRanges

    !-------------------------------------------------------------------
    ! Start

    numComps = dssComponents(dssPointer)
    if (numComps .le. 0) then
      dssRangeSpan%lower%type = UNDEFINED
      dssRangeSpan%lower%value = 0.0
      dssRangeSpan%upper%type = UNDEFINED
      dssRangeSpan%upper%value = 0.0
  return
    end if

    dssComp = dssComponent(dssPointer, 0)
    call readRangesComp(dssComp, axisName, oredRanges)

    do compNum = 1, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)
      call readRangesComp(dssComp, axisName, compRanges)
!      tempOredRanges => orRangesPair(compRanges, oredRanges)
!      deallocate(oredRanges) ! clears the memory
!      oredRanges => tempOredRanges
call orRanges(compRanges, oredRanges)
deallocate(compRanges)
    end do

    numRanges = size(oredRanges)
    dssRangeSpan%lower%type  = oredRanges(1)%lower%type
    dssRangeSpan%lower%value = oredRanges(1)%lower%value
    dssRangeSpan%upper%type  = oredRanges(numRanges)%upper%type
    dssRangeSpan%upper%value = oredRanges(numRanges)%upper%value

    deallocate(oredRanges) ! also clears tempOredRanges, since they
    ! are associated and thus point to the same memory.

  end function getRangesSpan

!***** add a variant that takes just a dss component??



  !---------------------------------------------------------------------
  subroutine readRangesDss(dssPointer, axisName, ranges, dssConstraints)
  !---------------------------------------------------------------------
  ! The function of this subroutine is to return a vector of dss type
  ! RangeT containing the ranges of the axis 'axisName'. Note that the sub
  ! will succeed only if that subset of the set of DSS components which obeys the supplied constraints can be expressed as (ranges in axisName) OR (all filters on other axes). This has been a complicated job to implement because a chain of ranges is implicitly an OR of the individual ranges and for this reason can be distributed among several components. The subroutine works therefore as follows. Firstly, blocks of components are identified which each include all components obeying the constraints and having the same range chain in axisName. Secondly, the non-axisName filters of each block of components are checked for identity. If they all prove to be the same, then their respective range chains are ORed together.
  !
  ! The returned pointer ranges should be deallocated after use.

    implicit none

    type(DssT),              intent(in)            :: dssPointer
    character(*),            intent(in)            :: axisName
    type(RangeT),            pointer               :: ranges(:)
    type(dssConstraintType), intent(in), optional  :: dssConstraints(:)

    logical :: thereAreConstraints, noCompObeysConstraints, isNewRangeChain
    integer :: numComps, compNum, numCompSubsets, compSubsetNum
    type(dssCompSubsetType), pointer :: compSubsets(:) => null()
    type(DScompT) :: dssComp
    type(RangeT), pointer :: compRanges(:) => null()
!, tempRanges(:)
    character(dssFiltCharLength) :: axisNamesToNeglect(1)

    !-------------------------------------------------------------------
    ! Start

    thereAreConstraints = .false.
    if (present(dssConstraints)) then
      if (size(dssConstraints) > 0) thereAreConstraints = .true.
    end if

    numComps = dssComponents(dssPointer)
    if (numComps .le. 0) then
      allocate(ranges(0))
  return
    end if

    noCompObeysConstraints = .true. ! default
    do compNum = 0, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)

      if (thereAreConstraints) then
    if (.not. dssCompObeys(dssComp, dssConstraints)) cycle
      end if

      noCompObeysConstraints = .false.
    exit
    end do

    if (noCompObeysConstraints) then
      allocate(ranges(0))
  return
    end if

    !...................................................................
    ! Start of main routine. First load all the blocks of constraint-
    ! -passing components:

    numCompSubsets = 0
    allocate(compSubsets(0))
    do compNum = 0, numComps - 1
      dssComp = dssComponent(dssPointer, compNum)

      if (thereAreConstraints) then
    if (.not. dssCompObeys(dssComp, dssConstraints)) cycle
      end if

      call readRangesComp(dssComp, axisName, compRanges)

      ! Does it have a range chain the same as any of the previous range chains?
      isNewRangeChain = .true. ! default
      do compSubsetNum = 1, numCompSubsets	! ignored in the 1st visit
        if (rangesPairIsIdentical(compRanges&
          , compSubsets(compSubsetNum)%ranges)) then

          isNewRangeChain = .false.
        end if
      end do

      if (isNewRangeChain) then ! make a new component subset:
        call addNewCompSubset(compSubsets, numCompSubsets, compRanges, compNum)
      else ! it is the same ranges as a previous block
        call addNewCompNumToSubset(compSubsets(numCompSubsets)%compNums&
          , compNum)
      end if

      deallocate(compRanges)
    end do
if (numCompSubsets <= 0) call error('dss_auxBug3',"ssclib:readRangesDss: please contact the code developer.")

    !...................................................................
    ! Now check that, apart from filters on axisName, the subsets are the same:

    axisNamesToNeglect(1) = axisName
    do compSubsetNum = 2, numCompSubsets
      if (.not. compSubsetsIdentical(dssPointer, compSubsets(compSubsetNum)&
        , compSubsets(1), axisNamesToNeglect)) then

        allocate(ranges(0))
	call deallocateDssCompSubsets(compSubsets)	! deallocate all compSubsets() (MAS)
        !do iLong=1, size(compSubsets)	! deallocate all compSubsets() (MAS)
        !  call deallocateDssCompSubset(compSubsets(iLong))
        !end do
        deallocate(compSubsets)
  return
      end if
    end do

    !...................................................................
    ! OR together the total range chain:

    allocate(ranges(size(compSubsets(1)%ranges)))
    ranges = compSubsets(1)%ranges
    do compSubsetNum = 2, numCompSubsets
!      tempRanges => orRangesPair(compSubsets(compSubsetNum)%ranges, ranges)
!      deallocate(ranges)
!      ranges => tempRanges
call orRanges(compSubsets(compSubsetNum)%ranges, ranges)
    end do

    !...................................................................
    ! Clear up all the subset pointers:

    if ( numCompSubsets .ne. size(compSubsets) ) &
       call error('bugCompSubSets', 'Contact the code developper')
    do compSubsetNum = 1, numCompSubsets
      deallocate(compSubsets(compSubsetNum)%compNums, compSubsets(compSubsetNum)%ranges)	! should be replaced by call deallocateDssCompSubset(compSubsets)
    end do
    deallocate(compSubsets)

  end subroutine readRangesDss


  !---------------------------------------------------------------------
  subroutine deallocateDssConstraint(dssConstraint)
  !---------------------------------------------------------------------

    implicit none

    type(dssConstraintType), intent(inout) :: dssConstraint

    !-------------------------------------------------------------------
    ! Start

    deallocate(dssConstraint%ranges)

  end subroutine deallocateDssConstraint

  !---------------------------------------------------------------------
  subroutine deallocateDssCompSubset(dssCompSubset)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompSubsetType), intent(inout) :: dssCompSubset

    !-------------------------------------------------------------------
    ! Start

    deallocate(dssCompSubset%ranges)
    deallocate(dssCompSubset%compNums)

  end subroutine deallocateDssCompSubset


  !---------------------------------------------------------------------
  subroutine deallocateDssCompSubsetOne(dssCompSubset)
  !---------------------------------------------------------------------
  ! Duplicate of deallocateDssCompSubset()

    implicit none

    type(dssCompSubsetType), intent(inout) :: dssCompSubset

    call deallocateDssCompSubset(dssCompSubset)

  end subroutine deallocateDssCompSubsetOne


  !---------------------------------------------------------------------
  subroutine deallocateDssCompSubsetAry(dssCompSubsets)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompSubsetType), intent(inout) :: dssCompSubsets(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    do i=1, size(dssCompSubsets)
      call deallocateDssCompSubset(dssCompSubsets(i))
    end do

  end subroutine deallocateDssCompSubsetAry


  !---------------------------------------------------------------------
  subroutine addNewCompSubset(compSubsets, numCompSubsets, ranges, compNum)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompSubsetType), pointer   :: compSubsets(:)
    integer,             intent(inout) :: numCompSubsets
    type(RangeT),        intent(in)    :: ranges(:)
    integer,             intent(in)    :: compNum

    type(dssCompSubsetType), pointer :: tempCompSubsets(:) => null()
    integer :: i

    !-------------------------------------------------------------------
    ! Start

    allocate(tempCompSubsets(numCompSubsets))
    call copyCompSubsets(compSubsets, tempCompSubsets)
    do i = lbound(compSubsets, 1), ubound(compSubsets, 1)
      call deallocateDssCompSubset(compSubsets(i))
    end do
    deallocate(compSubsets)

    numCompSubsets = numCompSubsets + 1
    allocate(compSubsets(numCompSubsets))

    call copyCompSubsets(tempCompSubsets&
      , compSubsets(:numCompSubsets-1))
    do i = lbound(tempCompSubsets, 1), ubound(tempCompSubsets, 1)
      call deallocateDssCompSubset(tempCompSubsets(i))
    end do
    deallocate(tempCompSubsets)

    call copyRanges(ranges, compSubsets(numCompSubsets)%ranges)
    allocate(compSubsets(numCompSubsets)%compNums(1))
    compSubsets(numCompSubsets)%compNums(1) = compNum

  end subroutine addNewCompSubset

  !---------------------------------------------------------------------
  subroutine addNewCompNumToSubset(compNums, compNum)
  !---------------------------------------------------------------------
  ! Increse the array size of compNums by 1, with the component compNum
  ! Note: If you change this, then change the subroutine in test/testdssaux.f90
  !  with the same name.

    implicit none

    integer, pointer    :: compNums(:)
    integer, intent(in) :: compNum

    integer :: numComps
    integer, allocatable :: tempCompNums(:)

    !-------------------------------------------------------------------
    ! Start

    numComps = size(compNums)
    allocate(tempCompNums(numComps))
    tempCompNums = compNums
    deallocate(compNums)
    numComps = numComps + 1
    allocate(compNums(numComps))
    compNums(:numComps-1) = tempCompNums
    compNums(numComps) = compNum

  end subroutine addNewCompNumToSubset

  !---------------------------------------------------------------------
  subroutine copyCompSubsets(inCompSubsets, outCompSubsets)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompSubsetType), intent(in)  :: inCompSubsets(:)
    type(dssCompSubsetType), intent(out) :: outCompSubsets(size(inCompSubsets))

    integer :: numCompSubsets, compSubsetNum

    !-------------------------------------------------------------------
    ! Start

    numCompSubsets = size(inCompSubsets)
    do compSubsetNum = 1, numCompSubsets
      outCompSubsets(compSubsetNum)&
        = copyCompSubset(inCompSubsets(compSubsetNum))
    end do

  end subroutine copyCompSubsets

  !---------------------------------------------------------------------
  function copyCompSubset(inCompSubset) result(outCompSubset)
  !---------------------------------------------------------------------

    implicit none

    type(dssCompSubsetType), intent(in) :: inCompSubset
    type(dssCompSubsetType)             :: outCompSubset

    !-------------------------------------------------------------------
    ! Start

    call copyRanges(inCompSubset%ranges, outCompSubset%ranges)
    allocate(outCompSubset%compNums(size(inCompSubset%compNums)))
    outCompSubset%compNums = inCompSubset%compNums

  end function copyCompSubset

  !---------------------------------------------------------------------
  logical(bool) function compSubsetsIdentical(dssPointer, compSubsetA&
  , compSubsetB, axisNamesToNeglect)
  !---------------------------------------------------------------------
!**** Not finished. axisNamesToNeglect should be optional (also should include optional axisNamesToInclude), with some (graceful hopefully!) way of passing them to dssCompsIdentical().

    implicit none

    type(DssT),              intent(in) :: dssPointer
    type(dssCompSubsetType), intent(in) :: compSubsetA,&
                                           compSubsetB
    character(*),            intent(in) :: axisNamesToNeglect(:)

    integer :: numComps, compNumA, compNumB
    logical, allocatable :: compSubsetBCompMatched(:)
    logical :: unableToMatchComponent

    !-------------------------------------------------------------------
    ! Start

    compSubsetsIdentical = .false. ! default

    numComps = size(compSubsetA%compNums)
  if (numComps .ne. size(compSubsetB%compNums)) return

    allocate(compSubsetBCompMatched(numComps))
    compSubsetBCompMatched = .false. ! initially.
    do compNumA = 1, numComps
      unableToMatchComponent = .true. ! default
      do compNumB = 1, numComps
      if (compSubsetBCompMatched(compNumB)) cycle
        if (dssCompsIdentical(dssPointer, compSubsetA%compNums(compNumA)&
          , compSubsetB%compNums(compNumB), axisNamesToNeglect)) then
          unableToMatchComponent = .false.
          compSubsetBCompMatched(compNumB) = .true.
      exit
        end if
      end do

  if (unableToMatchComponent) return
    end do

  if (any(.not. compSubsetBCompMatched)) return

    compSubsetsIdentical = .true. ! made it through all the tests.

  end function compSubsetsIdentical

  !---------------------------------------------------------------------
  subroutine readRangesComp(dssComp, axisName, compRanges)
  !---------------------------------------------------------------------
  ! NOTE that the pointer compRanges() should be deallocated after use.

    type(DScompT), intent(in) :: dssComp
    character(*),  intent(in) :: axisName
    type(RangeT), pointer     :: compRanges(:)	! intent(out)

    integer :: filtNum, nSize
    type(DSfilterT) :: filter
    type(RangeT),  pointer :: ranges(:) => null()
!, tempRangeChain(:)
    character(dssFiltCharLength) :: dssFilterTypeStr, tmpDssFiltName

    !-------------------------------------------------------------------
    ! Start

    allocate(compRanges(1))
    compRanges%lower%type = UNDEFINED
    compRanges%upper%type = UNDEFINED
    compRanges%lower%value = 0.0 ! unnecessary, but I don't
    compRanges%upper%value = 0.0 ! like to leave them hanging.

  if (numFiltersOfAxis(dssComp, axisName) .le. 0) return

    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
      tmpDssFiltName = dssFilterName(filter)
    if (trim(tmpDssFiltName) .ne. axisName) cycle
    !if (trim(dssFilterName(filter)) .ne. axisName) cycle

      if (dssFilterType(filter) .eq. RANGEFILTERT) then
        call dssFilterRangesSize(filter,nSize)
      	allocate(ranges(0:nSize-1))
        ranges = dssRanges(filter)
!        tempRangeChain => andRangesPair(compRanges, ranges)
!        deallocate(compRanges)
!        deallocate(ranges)
!        compRanges => tempRangeChain
call andRanges(ranges, compRanges)
deallocate(ranges)

      else
        allocate(compRanges(0))
        dssFilterTypeStr = filterTypeToStr(dssFilterType(filter))
        call warning('badDssFilterType', "ssclib:readRangesComp: A filter of type "&
          //trim(dssFilterTypeStr)//" on the axis "//trim(axisName)&
          //" was detected. I couldn't read the information from this: &
          &I can only read the information in RANGEFILTERT filters.")
      end if
!nullify(filter%ptr)
    end do

  end subroutine readRangesComp

  !---------------------------------------------------------------------
  subroutine readGtisFromDssBlock(set, inBlock, ccdNum, gti)
  !---------------------------------------------------------------------
  ! NOTE! The pointer gti() should be deallocated after use.

    implicit none

    type(DataSetT),  intent(in) :: set
    type(BlockT),    intent(in) :: inBlock
    integer,         intent(in) :: ccdNum
    type(IntervalT), pointer    :: gti(:)	! intent(out), IntervalT defined in caltypes

    type(DssT)      :: dssPointer
    type(DScompT)   :: dssComp
    integer         :: compNum, filtNum, filterType, foreignStatus
    type(DSfilterT) :: filter
    logical         :: firstFilterOnTime
    ! type(IntervalT), pointer :: localGti(:) => null()
    type(IntervalT), pointer :: localGti(:)
    character(dssFiltCharLength) :: tmpDssFiltName
    integer :: i
logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    localGti => null()
    dssPointer = dataSubspace(inBlock)

    if (numDssCompsWithValue(dssPointer, 'CCDNR', real(ccdNum, single))<=0) then
      write(errStr, '(a,i2.2)') "ssclib:readGtisFromDssBlock: &
         &failed to find a DSS component for CCD=", ccdNum
      call message(appMsg, Noisy)
      allocate(gti(0))
      call dssRelease(dssPointer)
  return
    end if

if (testFlag) then 	!! DEBUG
    do i=0, dssComponents(dssPointer) - 1
      dssComp = dssComponent(dssPointer, i)
      print *,'DEBUG: dssComp No = ', i
      call dumpDssComponent(dssComp)
    end do
end if

    ! Find the dss component with the correct CCDNR:
    call getNum1stDssCompWithValue(dssPointer, 'CCDNR', real(ccdNum, single)&
       , compNum, filterType=RANGEFILTERT, status=foreignStatus)
    ! The selection of the CCDNR is (by DSSLib) based on RANGEFILTERT,
    ! (which is defined in dsslib/src/dsscore/dssinter.f90).
    ! What you want here is supposedly the mere GTIs for a particular CCDNR,
    ! whatever the other conditions, such as the pattern selection, are.
    ! Therefore filterType=RANGEFILTERT is specified rather than left blank
    ! (in which case all the other filters have the same condition to match).

    if (foreignStatus < 0) then
      allocate(gti(0))
      call dssRelease(dssPointer)
  return
    end if

    dssComp = dssComponent(dssPointer, compNum)
!     if (testFlag) then 
!       call dumpDssComponent(dssComp)
!     end if

    if (numFiltersOfAxis(dssComp, 'TIME') <= 0) then
      allocate(gti(0))
      call dssRelease(dssPointer)
  return
    end if

    firstFilterOnTime = .true.
    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
      tmpDssFiltName = dssFilterName(filter)
    if (trim(tmpDssFiltName) .ne. 'TIME') cycle
    !if (trim(dssFilterName(filter)) .ne. 'TIME') cycle

!      if (dssFilterType(filter) .eq. GTIFILTERT) then
      filterType = dssFilterType(filter)
      if (filterType .eq. GTIFILTERT) then
        if (firstFilterOnTime) then
          firstFilterOnTime = .false.
          call getGtiFromFilter(set, filter, gti)

        else
          call getGtiFromFilter(set, filter, localGti)
          call andIntervals(localGti, gti)
          deallocate(localGti)
        end if

      else
        write(errStr, '(a, i2.2, a, i3.3, a, i3.3, a)') "ssclib:readGtisFromDssBlock: &
          &This routine attempts to read a DSS filter on the TIME column for CCD "&
          , ccdNum, ". A filter of type ID ", filterType, " was found, however only &
          &filters of type GTIFILTERT (ID=", GTIFILTERT, ") are recognized."
        call warning('dssUnknownTimeFilter')
      end if
    end do

    call dssRelease(dssPointer)

  end subroutine readGtisFromDssBlock

  !---------------------------------------------------------------------
  subroutine readGtisFromDssPointer(set, dssPointer, ccdNum, gtis)
  !---------------------------------------------------------------------
  ! NOTE! The pointer gtis() should be deallocated after use.

    implicit none

    type(DataSetT),  intent(in) :: set
    type(DssT),      intent(in) :: dssPointer
    integer,         intent(in) :: ccdNum
    type(IntervalT), pointer    :: gtis(:)	! intent(out), IntervalT defined in caltypes

    type(DScompT)   :: dssComp
    integer         :: compNum, filtNum, filterType, foreignStatus
    type(DSfilterT) :: filter
    logical         :: firstFilterOnTime
    type(IntervalT), pointer :: localGtis(:) => null()
    character(dssFiltCharLength) :: tmpDssFiltName

    !-------------------------------------------------------------------
    ! Start

    if (numDssCompsWithValue(dssPointer, 'CCDNR', real(ccdNum, single))<=0) then
      write(errStr, '(a,i2.2)') "ssclib:readGtisFromDssPointer: &
         &failed to find a DSS component for CCD=", ccdNum
      call message(appMsg, Noisy)
      allocate(gtis(0))
  return
    end if

    ! Find the dss component with the correct CCDNR:
    call getNum1stDssCompWithValue(dssPointer, 'CCDNR', real(ccdNum, single)&
       , compNum, filterType=RANGEFILTERT, status=foreignStatus)
    ! See the comment in readGtisFromDssBlock() for detail.

    if (foreignStatus < 0) then
      allocate(gtis(0))
  return
    end if

    dssComp = dssComponent(dssPointer, compNum)

    if (numFiltersOfAxis(dssComp, 'TIME') <= 0) then
      allocate(gtis(0))
  return
    end if

    firstFilterOnTime = .true.
    do filtNum = 0, dssFilters(dssComp) - 1
      filter = dssFilter(dssComp, filtNum)
      tmpDssFiltName = dssFilterName(filter)
    if (trim(tmpDssFiltName) .ne. 'TIME') cycle
    !if (trim(dssFilterName(filter)) .ne. 'TIME') cycle

!      if (dssFilterType(filter) .eq. GTIFILTERT) then
      filterType = dssFilterType(filter)
      if (filterType .eq. GTIFILTERT) then
        if (firstFilterOnTime) then
          firstFilterOnTime = .false.
          call getGtiFromFilter(set, filter, gtis)

        else
          call getGtiFromFilter(set, filter, localGtis)
          call andIntervals(localGtis, gtis)
          deallocate(localGtis)
        end if

      else
        write(errStr, '(a, i2.2, a, i3.3, a, i3.3, a)') "ssclib:readGtisFromDssPointer: &
          &This routine attempts to read a DSS filter on the TIME column for CCD "&
          , ccdNum, ". A filter of type ID ", filterType, " was found, however only &
          &filters of type GTIFILTERT (ID=", GTIFILTERT, ") are recognized."
        call warning('dssUnknownTimeFilter')
      end if
    end do

  end subroutine readGtisFromDssPointer

  !---------------------------------------------------------------------
  subroutine getGtiFromFilter(set, filter, gti)
  !---------------------------------------------------------------------
  ! NOTE! The pointer gti() should be deallocated after use.

    implicit none

    type(DataSetT),  intent(in) :: set
    type(DSfilterT), intent(in) :: filter
    type(IntervalT), pointer    :: gti(:)	! intent(out), IntervalT defined in caltypes

    character(dssFilterBlockCharLength)  :: gtiBlockName
    type(TableT)    :: tab
    integer         :: numGtis

    !-------------------------------------------------------------------
    ! Start

    gtiBlockName = ''
    gtiBlockName = dssFilterBlock(filter)
    ! gtiBlockName = trim(dssFilterBlock(filter))
    tab = table(set, trim(gtiBlockName))
    numGtis = numberOfRows(tab)

    write(errStr, '(a,i8)') "ssclib:getGtiFromFilter: &
       &Reading ("//trim(gtiBlockName)//").  Number of Rows = ", numGtis
    call message(appMsg, Noisy)

    allocate(gti(numGtis))
    call readColDataToFixed(tab, 'START', gti%lower)
    call readColDataToFixed(tab, 'STOP',  gti%upper)
    call release(tab)

  end subroutine getGtiFromFilter

  !---------------------------------------------------------------------
  subroutine decomposeRegionXyNameString(xyNameString, xColName, yColName)
  !---------------------------------------------------------------------
  ! Designed to analyse the output of the call dssFilterComps, which returns a string containing the names of the 'X' and 'Y' columns of a region specification block separated by a comma.

    implicit none

    character(*), intent(in) :: xyNameString
    character(*), intent(out) :: xColName, yColName

    integer :: posOfComma

    !-------------------------------------------------------------------
    ! Start

    xColName = '' ! default
    yColName = '' ! default

    posOfComma = index(xyNameString, ',')
    if (posOfComma == 0)&
call error('commaNotFound', 'ssclib:decomposeRegionXyNameString: The call &
&dssFilterComps is expected to return a string containing the names of the &
&X and Y columns, separated by a comma. However no comma was found in the string.')

    xColName = trim(adjustl(xyNameString(:posOfComma-1)))
    yColName = trim(adjustl(xyNameString(posOfComma+1:)))

  end subroutine decomposeRegionXyNameString

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module dss_aux
