! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: polygon
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains several routines to do with dividing up polygons on a square grid.
! 
! Comments: TODO:
! 
! $Id: coordinate_mod.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module polygon

  use types

  use constants
  use geometric_types	! for Point2dT type
  use ssc_misc		! for isOdd, fracVal

  implicit none 

  private 

  public areaOfPolygon
  public areaOfPolyCurve
  public isConvex
  public pointIsInsidePoly
  public copyCurvedSide
  public initCurvedSide
  public deallocatePolycurveVectorT

  type, public :: CurvedSideT ! appropriate for a 'polygon' made up from
  ! a combination of straight lines and arcs.
    type(Point2dT) :: pos, centre
    logical(bool)  :: isStraight, arcIsClockwise, centreLiesToRight
    real(single)   :: angle, radius
    integer(int8)  :: field
  end type CurvedSideT

  type, public :: PolycurveVectorT
    type(CurvedSideT), pointer :: poly(:) => null()
  end type PolycurveVectorT

  integer, parameter, public :: INSIDE_BDRY=0, ON_BDRY=1, OUTSIDE_BDRY=2
contains 

  !---------------------------------------------------------------------
  function areaOfPolygon(poly) result(area)
  !---------------------------------------------------------------------
  ! Surprise surprise, calculates the area of the given polygon. The basic
  ! formula is derived (not by me, I got it off the comp.graphics.algorithms
  ! FAQ, but it is pretty obvious) by considering the polygon as two curves
  ! on the x-y plane. Each curve is a sequence of line segments and the
  ! curves are joined at both low-x and hi-x extrema. This is a definition
  ! for a polygon, right?? Now the area of the polygon can be easily seen
  ! to be equal to the integral of the lower curve subtracted from that of
  ! the upper curve. But the integral under a line segment from (x0,y0) to
  ! (x1,y1) is just
  !
  !	A = 0.5 (x1 - x0) * (y1 + y0).
  !
  ! Note that this is negative if x1 < x0. The total area of the polygon Ap
  ! can thus be expressed as the sum of the integrals of all its line
  ! segments: the 'bottom' line segments contribute the necessary negative
  ! factor. Thus
  !
  !	Ap = 0.5 sum_{i=1}^{N} ((x_i - x_{i-1}) * (y_i + y_{i-1}))
  !
  ! where 'i-1' = N if i = 1. This can be boiled down to
  !
  !	Ap = 0.5 sum_{i=1}^{N} (x_i * (y_{i-1} - y_{i+1}))
  !
  ! which involves fewer operations. Note that the answer is negative if
  ! the points in the polynomial are numbered in an anticlockwise sense.

    implicit none

    type(Point2dT), intent(in) :: poly(:)
    real(single) :: area

    integer :: i, numPoints

    !-------------------------------------------------------------------
    ! Start

    area = 0.0
    numPoints = size(poly)
  if (numPoints <= 2) return

    area = area + poly(1)%x * (poly(numPoints)%y - poly(2)%y)
    do i = 2, numPoints - 1
      area = area + poly(i)%x * (poly(i - 1)%y - poly(i + 1)%y)
    end do
    area = area + poly(numPoints)%x * (poly(numPoints - 1)%y - poly(1)%y)
    area = area / 2.0

  end function areaOfPolygon

  !---------------------------------------------------------------------
  function areaOfPolyCurve(polyCurve) result(area)
  !---------------------------------------------------------------------

    implicit none

    type(CurvedSideT), intent(in) :: polyCurve(:)
    real(single) :: area

    integer :: i, iNext, numPts

    !-------------------------------------------------------------------
    ! Start

    area = areaOfPolygon(polyCurve%pos)

    ! Now add (or subtract) the bits due to curved sides:

    numPts = size(polyCurve)
    do i = 1, numPts
    if (polyCurve(i)%isStraight) cycle
      iNext = 1 + modulo(i, numPts)

      area = area + additionFromCurve(polyCurve(i)%pos, polyCurve(iNext)%pos&
        , polyCurve(i)%radius, polyCurve(i)%arcIsClockwise&
        , polyCurve(i)%centreLiesToRight)
    end do

  end function areaOfPolyCurve

  !---------------------------------------------------------------------
  function additionFromCurve(loPos, hiPos, radius, isCw, onRight) result(area)
  !---------------------------------------------------------------------

    implicit none

    type(Point2DT), intent(in) :: loPos, hiPos
    real(single),   intent(in) :: radius
    logical(bool),  intent(in) :: isCw, onRight
    real(single)               :: area

    type(Point2DT) :: posDiff
    real(single) :: sideLength, halfAngle, d, triangleArea

    !-------------------------------------------------------------------
    ! Start

    posDiff%x = hiPos%x - loPos%x
    posDiff%y = hiPos%y - loPos%y
    sideLength = sqrt(posDiff%x*posDiff%x+posDiff%y*posDiff%y)
    halfAngle = asin(sideLength/2.0/radius)
    d = radius * cos(halfAngle)
    triangleArea = d * sideLength / 2.0

    if (isCw) then
      if (onRight) then
        area =        halfAngle  * radius * radius - triangleArea
      else
        area =  (Pi - halfAngle) * radius * radius + triangleArea
      end if
    else ! acw
      if (onRight) then
        area = -(Pi - halfAngle) * radius * radius - triangleArea
      else
        area =       -halfAngle  * radius * radius + triangleArea
      end if
    end if

  end function additionFromCurve

  !---------------------------------------------------------------------
  logical(bool) function isConvex(pos0, pos1, pos2)
  !---------------------------------------------------------------------

    implicit none

    type(Point2dT), intent(in) :: pos0, pos1, pos2

    type(Point2dT) :: vectorA, vectorB

    !-------------------------------------------------------------------
    ! Start

    if ( (pos0%x .eq. pos1%x .and. pos0%y .eq. pos1%y)&
    .or. (pos0%x .eq. pos2%x .and. pos0%y .eq. pos2%y)&
    .or. (pos1%x .eq. pos2%x .and. pos1%y .eq. pos2%y)) then
      isConvex = .true. ! why not? Neither answer is sensible in these cases.

    else
      vectorA%x = pos1%x - pos0%x
      vectorA%y = pos1%y - pos0%y
      vectorB%x = pos2%x - pos1%x
      vectorB%y = pos2%y - pos1%y
      if (vectorA%x * vectorB%y < vectorA%y * vectorB%x) then
        isConvex = .true.
      else
        isConvex = .false.
      end if
    end if

  end function isConvex

  !---------------------------------------------------------------------
  logical(bool) function pointIsInsidePoly(x, y, poly, includeBorder)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in)           :: x, y
    type(Point2dT), intent(in)           :: poly(:)
    logical,        intent(in), optional :: includeBorder

    logical(bool) :: excludeBorder
    real(single) :: xFrac
    real(single), allocatable :: intersectionY(:)
    integer :: numIntersections, i, numHigherIntersections

    !-------------------------------------------------------------------
    ! Start

    allocate(intersectionY(size(poly)))

    if (present(includeBorder)) then
      excludeBorder = .not. includeBorder
    else
      excludeBorder = .true. ! ie point is stated to be inside the poly only
      ! if it is neither outside nor on the border. In other words, the
      ! border is considered to be outside the poly.
    end if

    ! Calculate the y values of all points of intersection between polygon
    ! sides and a vertical line passing through x:
    numIntersections = 0
    do i = 1, size(poly) - 1
      if (poly(i)%x == x .and. poly(i)%y == y) then
        if (excludeBorder) then
          pointIsInsidePoly = .false.
        else
          pointIsInsidePoly = .true.
        end if
  return
      end if

      if (poly(i)%x <= x .and. poly(i + 1)%x > x) then
        numIntersections = numIntersections + 1
        xFrac = (x - poly(i)%x) / (poly(i + 1)%x - poly(i)%x)
        intersectionY(numIntersections)&
        = fracVal(poly(i)%y, poly(i + 1)%y, xFrac)
      end if
    end do

    if (numIntersections .le. 0) then
      pointIsInsidePoly = .false.
  return
    end if

    if (any(intersectionY(1:numIntersections) == y)) then
      if (excludeBorder) then
        pointIsInsidePoly = .false.
      else
        pointIsInsidePoly = .true.
      end if
  return
    end if

    ! Now find the number of intersections that have y value greater than y.
    ! If this is even, then (x,y) lies outside the poly; if odd, (x,y) lies
    ! within.
    numHigherIntersections = count(intersectionY(1:numIntersections) > y)
    if (isOdd(numHigherIntersections)) then
      pointIsInsidePoly = .true.
    else
      pointIsInsidePoly = .false.
    end if

  end function pointIsInsidePoly

  !---------------------------------------------------------------------
  subroutine copyCurvedSide(inPoly, outPoly)
  !---------------------------------------------------------------------

    implicit none 

    type(CurvedSideT), intent(in) :: inPoly(:)
    type(CurvedSideT), pointer    :: outPoly(:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(outPoly(size(inPoly)))

    outPoly%pos%x             = inPoly%pos%x
    outPoly%pos%y             = inPoly%pos%y
    outPoly%centre%x          = inPoly%centre%x
    outPoly%centre%y          = inPoly%centre%y
    outPoly%isStraight        = inPoly%isStraight
    outPoly%arcIsClockwise    = inPoly%arcIsClockwise
    outPoly%centreLiesToRight = inPoly%centreLiesToRight
    outPoly%angle             = inPoly%angle
    outPoly%radius            = inPoly%radius
    outPoly%field             = inPoly%field

  end subroutine copyCurvedSide

  !---------------------------------------------------------------------
  subroutine initCurvedSide(outPoly)
  !---------------------------------------------------------------------

    implicit none 

    type(CurvedSideT), intent(out) :: outPoly(:)

    !-------------------------------------------------------------------
    ! Start:

    outPoly%pos%x = 0.0
    outPoly%pos%y = 0.0
    outPoly%centre%x = 0.0
    outPoly%centre%y = 0.0
    outPoly%isStraight = .true.
    outPoly%arcIsClockwise = .false.
    outPoly%centreLiesToRight = .true.
    outPoly%angle  = 0.0
    outPoly%radius = 0.0
    outPoly%field = OUTSIDE_BDRY

  end subroutine initCurvedSide

  !---------------------------------------------------------------------
  subroutine deallocatePolycurveVectorT(polyVec)
  !---------------------------------------------------------------------

    implicit none 

    type(PolycurveVectorT), pointer :: polyVec(:)

    integer :: polyNum

    !-------------------------------------------------------------------
    ! Start:

    do polyNum = lbound(polyVec,1), ubound(polyVec,1)
    ! do polyNum = 1, size(polyVec)
      deallocate(polyVec(polyNum)%poly)
    end do
    deallocate(polyVec)

  end subroutine deallocatePolycurveVectorT

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module polygon
