! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: warp_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some nonlinear transforms.
! 
! Comments: TODO:
! 
! $Id: warp_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module warp_utils

  use types		! For single, double, int16 etc type definitions.

  use array_utils	! for median.

  implicit none 

  private 

  public scaleBalance
  public warp
  public biWarp

contains 

  !---------------------------------------------------------------------
  subroutine scaleBalance(vector, warpScaledMedianToArg, status)
  !---------------------------------------------------------------------
  ! This sub takes a set of points which might be all very near the maximum or minimum value (or even in a later manifestation, clumped near the middle), and warps the scale so that the median value becomes equal to half-way between the max and min vals.

    implicit none

    real(single), intent(inout)         :: vector(:)
    real(single), intent(in),  optional :: warpScaledMedianToArg
    integer,      intent(out), optional :: status

    integer :: vsize, i, localStatus, foreignStatus
    real(single) :: maax, miin, spread, medianVal, scaledMedianVal, sag, x, warpScaledMedianTo

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    if (present(warpScaledMedianToArg)) then
      warpScaledMedianTo = warpScaledMedianToArg
    else
      warpScaledMedianTo = 0.5
    end if

    maax = maxval(vector)
    miin = minval(vector)
    spread = maax - miin
    if (spread .le. 0.0) then
      if (present(status)) status = -1
  return
    end if

    vsize = size(vector)
    if (vsize .le. 0) then
      if (present(status)) status = -2
  return
    end if

    medianVal = median(vector)
    scaledMedianVal = (medianVal - miin) / spread
    sag = (scaledMedianVal - warpScaledMedianTo)&
    / (warpScaledMedianTo * (2.0 * scaledMedianVal - 1.0) - scaledMedianVal)
    sag = min(1.0, max(0.0, sag))

    do i = 1, vsize
      x = min(1.0, max(0.0, (vector(i) - miin) / spread))
      call warp(sag, x, vector(i), foreignStatus)

      if (foreignStatus < 0) then
        if (present(status)) status = foreignStatus - 2
  return
      end if
    end do

    if (present(status)) status = localStatus

  end subroutine scaleBalance

  !---------------------------------------------------------------------
  subroutine warp(sag, x, y, status)
  !---------------------------------------------------------------------
  ! The action of this subroutine is to warp the x values (which should
  ! be between 0 and 1) either closer to 0 or closer to 1, depending on
  ! the value of 'sag'. For sag = -1, the 'upward' warp is at its limit:
  ! whatever the value of x, y will be set to 1. For sag = +1 the warp is
  ! at the opposite extreme, in which the routine returns 0 for any value
  ! of x. For sag = 0, y = x. The interesting things happen when sag !=
  ! either -1, 0 or +1. In these cases the relation of y to x is a
  ! hyperbolic function that generates the required warp.
  !
  ! This subroutine gives accurate results except for x values close to 1.
  ! A call such as
  !
  !	call warp(-sag, 1.0-x, invY)
  !	y = 1.0 - invY
  !
  ! is better in this instance.

    implicit none

    real(single),    intent(in)            :: sag, x
    real(single),    intent(out)           :: y
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start:

    if (present(status)) status = 0 ! default

    if (x .gt. 1.0 .or. x .lt. 0.0) then
      if (present(status)) status = -1
  return
    end if

    if (abs(sag) .gt. 1.0) then
      if (present(status)) status = -2
  return
    end if

    if (sag .eq. -1.0) then
      y = 0.0
    else if (sag .eq. 1.0) then
      y = 1.0
    else
      y = x * (1.0 + sag) / (1.0 + sag * (2.0 * x - 1.0))
    end if

  end subroutine warp

  !---------------------------------------------------------------------
  subroutine biWarp(stretch, x, y, status)
  !---------------------------------------------------------------------
  ! The action of this subroutine is to stretch or compress the x
  ! values (which should be between -1 and +1) about 0, depending on the
  ! value of 'stretch'. For stretch = -1, all the points are compressed
  ! into one: whatever the value of x, y will be set to 0. For stretch
  ! = +1 the spread is at its maximum, in which the routine returns -1
  ! for negative values of x and +1 for positive values. At stretch = 0,
  ! y = x. The interesting things happen when stretch != either -1, 0
  ! or +1. In these cases the relation of y to x is a double hyperbolic
  ! function (generated by the appropriate call to the subroutine warp)
  ! that generates the required stretch.

    implicit none

    real(single),    intent(in)            :: stretch, x
    real(single),    intent(out)           :: y
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start:

    if (present(status)) status = 0 ! default

    if (abs(x) > 1.0) then
      if (present(status)) status = -1
  return
    end if

    if (abs(stretch) > 1.0) then
      if (present(status)) status = -2
  return
    end if

    if (stretch == -1.0) then
      y = 0.0
    else if (stretch == 1.0) then
      y = real(sign(1.0, x))
    else
      if (x > 0.0) then
        call warp(-stretch, x, y)
      else
        call warp(-stretch, -x, y)
        y = -y
      end if
    end if

  end subroutine biWarp

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module warp_utils
